# These are default plotting handlers that work in some common training
#  scenarios, but won't work in every case:

import logging
import os

from typing import Dict, Optional, Callable, Sequence

import matplotlib.figure
import matplotlib.pyplot as plt
import matplotlib.cm
import numpy as np
import torch
from torch.nn import functional as F

from elektronn3.data.utils import squash01
from elektronn3.data.transforms import RemapTargetIDs

E3_CMAP: str = os.getenv('E3_CMAP')

logger = logging.getLogger('elektronn3log')


def get_cmap(out_channels: int):
    if E3_CMAP is not None:
        cmname = E3_CMAP
    # Else, use defaults:
    elif out_channels <= 10:
        cmname = 'tab10'
    elif out_channels <= 20:
        cmname = 'tab20'
    else:
        raise RuntimeError(
            f'Default cmaps only support up to 20 colors, which are not enough to label '
            f'{out_channels} different output channels.\nPlease set a different cmap '
            'with the E3_CMAP envvar.'
        )
    return matplotlib.cm.get_cmap(cmname, out_channels)


def plot_image(
        image: np.ndarray,
        overlay: Optional[np.ndarray] = None,
        overlay_alpha=0.5,
        cmap=None,
        colorbar=True,
        filename=None,
        vmin=None,
        vmax=None
) -> matplotlib.figure.Figure:
    """Plots a 2D image to a malplotlib figure.

    For gray-scale images, use ``cmap='gray'``.
    For label matrices (segmentation targets or class predictions),
    specify the global number of possible classes in ``out_channels``."""

    # Determine colormap and set discrete color values if needed.
    ticks = None
    ticklabels = None
    cmap_name = cmap if isinstance(cmap, str) else cmap.name
    if cmap_name in {E3_CMAP, 'tab10', 'tab20'}: # qualitative cmap
        ticks = np.linspace(0.5, vmax - 0.5, vmax) # 0.5 for centered ticks
        ticklabels = np.arange(vmax)

    fig, ax = plt.subplots(constrained_layout=True, figsize=(10, 10))
    if image.ndim == 3 and image.shape[0] == 1:
        image = image[0]
    if overlay is None:
        aximg = ax.imshow(image, cmap=cmap, vmin=vmin, vmax=vmax, interpolation='none')
    else:
        ax.imshow(image, cmap='gray')
        masked_overlay = np.ma.masked_where(overlay == 0, overlay)
        aximg = ax.imshow(masked_overlay, cmap=cmap, vmin=vmin, vmax=vmax, alpha=overlay_alpha, interpolation='none')
    if filename is not None:
        max_filename_length = 50  # Truncate long file names from the left
        if len(filename) > max_filename_length:
            filename = f'...{filename[-max_filename_length:]}'
        ax.set_title(filename)
    if colorbar:
        bar = fig.colorbar(aximg, ticks=ticks)
        if ticklabels is not None:
            bar.set_ticklabels(ticklabels)
        bar.solids.set(alpha=1) # otherwise uses image’s opacity
    return fig


def _get_batch2img_function(
        batch: np.ndarray,
        z_plane: Optional[int] = None
) -> Callable[[np.ndarray], np.ndarray]:
    """
    Defines ``batch2img`` function dynamically, depending on tensor shapes.

    ``batch2img`` slices a 4D or 5D tensor to (C, H, W) shape, moves it to
    host memory and converts it to a numpy array.
    By arbitrary choice, the first element of a batch is always taken here.
    In the 5D case, the D (depth) dimension is sliced at z_plane.

    This function is useful for plotting image samples during training.

    Args:
        batch: 4D or 5D tensor, used for shape analysis.
        z_plane: Index of the spatial plane where a 5D image tensor should
            be sliced. If not specified, this is automatically set to half
            the size of the D dimension.

    Returns:
        Function that slices a plottable 2D image out of a np.ndarray
        with batch and channel dimensions.
    """
    if batch.ndim == 5:  # (N, C, D, H, W)
        if z_plane is None:
            z_plane = batch.shape[2] // 2
        assert z_plane in range(batch.shape[2])
        return lambda x: x[0, :, z_plane]
    elif batch.ndim == 4:  # (N, C, H, W)
        return lambda x: x[0, :]
    elif batch.ndim == 2:  # (N, C) -> img2scalar
        return lambda x: x[None, ]  # (1, N, C)  -> Image will show N x C probabilities
    else:
        raise ValueError('Only 4D and 5D tensors are supported.')


def write_to_kzip(trainer: 'Trainer', pred_batch: np.ndarray) -> None:
    from knossos_utils import KnossosDataset
    ks = trainer.knossos_preview_config
    if isinstance(ks['dataset'], str):
        dataset_path = ks['dataset']
    else:
        dataset_path = ks['dataset'][0]
    ds = KnossosDataset(dataset_path)
    seg = pred_batch[0].swapaxes(0, 2)  # (N, D, H, W) -> (W, H, D)

    # Set movement are in k.zip
    area_min = ks['offset']
    area_sz = ks['size']
    anno_str = f"""<?xml version="1.0" encoding="UTF-8"?>
<things>
    <parameters>
        <MovementArea min.x="{area_min[0]}" min.y="{area_min[1]}" min.z="{area_min[2]}" size.x="{area_sz[0]}" size.y="{area_sz[1]}" size.z="{area_sz[2]}"/>
    </parameters>
    <comments/>
    <branchpoints/>
</things>"""

    kzip_path = f'{trainer.save_path}/preview_{trainer.step}.k.zip'
    logger.info(f'Writing preview inference to {kzip_path}')
    ds.save_to_kzip(
        data=seg,
        data_mag=ks.get('mag', 1),
        kzip_path=kzip_path,
        offset=ks['offset'],
        mags=ks.get('target_mags', [1, 2]),
        gen_mergelist=False,
        upsample=False,
        fast_resampling=False,
        annotation_str=anno_str
    )


# TODO: Support regression scenario
def _tb_log_preview(
        trainer: 'Trainer',  # For some reason Trainer can't be imported
        z_plane: Optional[int] = None,
        group: str = 'preview'
) -> None:
    """Preview from constant region of preview batch data."""
    inp_batch = trainer.preview_batch
    out_batch = trainer._preview_inference(
        inp=inp_batch,
        inference_kwargs=trainer.inference_kwargs,
    ).to(torch.float32)
    inp_batch = inp_batch.numpy()
    if trainer.inference_kwargs['apply_softmax']:
        out_batch = F.softmax(out_batch, 1).numpy()
    else:
        out_batch = out_batch.numpy()

    batch2img = _get_batch2img_function(out_batch, z_plane)

    if inp_batch.ndim == 5 and trainer.enable_videos:
        # 5D tensors -> 3D images -> We can make 2D videos out of them
        # See comments in the 5D section in _tb_log_sample_images
        inp_video = squash01(inp_batch)
        # (N, C, T=D, H, W) -> (N, T=D, C, H, W) because of add_video API
        inp_video = np.swapaxes(inp_video, 1, 2)
        trainer.tb.add_video(
            f'{group}_vid/inp', inp_video, global_step=trainer.step
        )
        for c in range(out_batch.shape[1]):
            outc_video = squash01(out_batch[:, c][None])  # Slice C, but keep dimensions intact
            # (N, C=1, T=D, H, W) -> (N, T=D, C=1, H, W)
            outc_video = np.moveaxis(outc_video, 1, 2)
            trainer.tb.add_video(
                f'{group}_vid/out{c}',
                outc_video,
                global_step=trainer.step
            )

    out_slice = batch2img(out_batch)
    pred_slice = out_slice.argmax(0)

    if trainer.knossos_preview_config is not None:
        pred_batch = out_batch.argmax(1)
        remap_ids = trainer.knossos_preview_config.get('remap_ids')
        if remap_ids is not None:
            remap = RemapTargetIDs(remap_ids, reverse=True)
            _, pred_batch = remap(None, pred_batch)

        write_to_kzip(trainer, pred_batch)

    for c in range(out_slice.shape[0]):
        trainer.tb.add_figure(
            f'{group}/out{c}',
            plot_image(out_slice[c], cmap='gray'),
            trainer.step
        )
    class_cmap = get_cmap(trainer.max_plot_id)
    trainer.tb.add_figure(
        f'{group}/pred',
        plot_image(pred_slice, vmin=0, vmax=trainer.max_plot_id, cmap=class_cmap),
        trainer.step
    )
    inp_slice = batch2img(inp_batch)[0]
    trainer.tb.add_figure(
        f'{group}/pred_overlay',
        plot_image(inp_slice, overlay=pred_slice, overlay_alpha=trainer.overlay_alpha, vmin=0, vmax=trainer.max_plot_id, cmap=class_cmap),
        global_step=trainer.step
    )

    # This is only run once per training, because the ground truth for
    # previews is constant (always the same preview inputs/targets)
    if trainer._first_plot:
        inp_slice = batch2img(trainer.preview_batch)[0]
        trainer.tb.add_figure(
            f'{group}/inp',
            plot_image(inp_slice, cmap='gray'),
            global_step=0
        )
        trainer._first_plot = False


def _tb_log_sample_images(
        trainer: 'Trainer',
        images: Dict[str, np.ndarray],
        z_plane: Optional[int] = None,
        group: str = 'sample'
) -> None:
    """Preview from last training/validation sample

    Since the images are chosen randomly from the training/validation set
    they come from random regions in the data set.

    Note: Training images are possibly augmented, so the plots may look
        distorted/weirdly colored.
    """
    # Always only use the first element of the batch dimension
    inp_batch = images['inp'][:1]
    target_batch = images.get('target')
    if target_batch is not None:
        target_batch = target_batch[:1]
    out_batch = images['out'][:1]

    name = images.get('fname')
    if name is not None:
        name = name[0]

    continuous_cmap = 'viridis'

    if trainer.inference_kwargs['apply_softmax']:
        out_batch = F.softmax(torch.as_tensor(out_batch, dtype=torch.float32), 1).numpy()
    elif trainer.inference_kwargs.get('apply_sigmoid'):
        out_batch = torch.sigmoid(torch.as_tensor(out_batch, dtype=torch.float32)).numpy()
    else:
        out_batch = out_batch.astype(np.float32)

    batch2img_inp = _get_batch2img_function(inp_batch, z_plane)
    inp_slice = batch2img_inp(images['inp'])

    uinp_batch = images.get('unlabeled')
    if uinp_batch is not None:
        trainer.tb.add_figure(
            f'{group}/unlabeled_inp',
            plot_image(batch2img_inp(uinp_batch['inp'].cpu().numpy())[0], cmap='gray'),
            global_step=trainer.step
        )

    # TODO: Support one-hot targets
    # TODO: Support multi-label targets
    # TODO: Output vis missing if target_batch is None
    # Check if the network is being trained for classification with class index target tensors
    if target_batch is not None:
        is_classification = target_batch.ndim == out_batch.ndim - 1
        class_cmap = get_cmap(trainer.max_plot_id)
        # If it's not classification, we assume a regression scenario
        is_regression = np.all(target_batch.shape == out_batch.shape)
        # If not exactly one of the scenarios is detected, we can't handle it
        assert is_regression != is_classification

        if is_classification:
            # In classification scenarios, targets have one dim less than network
            #  outputs, so if we want to use the same batch2img function for
            #  targets, we have to add an empty channel axis to it after the N dimension
            target_batch = target_batch[:, None]

    inp_sh = np.array(inp_batch.shape[2:])
    out_sh = np.array(out_batch.shape[2:])
    if out_batch.shape[2:] != inp_batch.shape[2:] and not (out_batch.ndim == 2):
        # Zero-pad output and target to match input shape
        # Create a central slice with the size of the output
        lo = (inp_sh - out_sh) // 2
        hi = inp_sh - lo
        slc = tuple([slice(None)] * 2 + [slice(l, h) for l, h in zip(lo, hi)])

        padded_out_batch = np.zeros(
            (inp_batch.shape[0], out_batch.shape[1], *inp_batch.shape[2:]),
            dtype=out_batch.dtype
        )
        padded_out_batch[slc] = out_batch
        out_batch = padded_out_batch

        # Assume that target has the same shape as the output and pad it, too
        if target_batch is not None:
            padded_target_batch = np.zeros((*target_batch.shape[:2], *inp_batch.shape[2:]), dtype=target_batch.dtype)
            padded_target_batch[slc] = target_batch
            target_batch = padded_target_batch

    target_cmap = E3_CMAP
    batch2img = _get_batch2img_function(out_batch, z_plane)
    if target_batch is not None:
        target_slice = batch2img(target_batch)
    out_slice = batch2img(out_batch)
    if target_batch is not None:
        if is_classification:
            target_slice = target_slice.squeeze(0)  # Squeeze empty axis that was added above
        elif target_slice.shape[0] == 3:  # Assume RGB values
            # RGB images need to be transposed to (H, W, C) layout so matplotlib can handle them
            target_slice = np.moveaxis(target_slice, 0, -1)  # (C, H, W) -> (H, W, C)
            out_slice = np.moveaxis(out_slice, 0, -1)

    if inp_batch.ndim == 5 and trainer.enable_videos:
        # 5D tensors -> 3D images -> We can make 2D videos out of them
        # We re-interpret the D dimension as the temporal dimension T of the video
        #  -> (N, T, C, H, W)
        # Inputs and outputs need to be squashed to the (0, 1) intensity range
        #  for video rendering, otherwise they will appear as random noise.
        # Since tensorboardX's add_video only supports (N, T, C, H, W) tensors,
        #  we have to add a fake C dimension to the (N, D, H, W) target tensors
        #  and replace the C dimension of output tensors by empty C dimensions
        #  to visualize each channel separately.
        inp_video = squash01(inp_batch)
        # (N, C, T=D, H, W) -> (N, T=D, C, H, W) because of add_video API
        inp_video = np.swapaxes(inp_video, 1, 2)
        trainer.tb.add_video(
            f'{group}_vid/inp', inp_video, global_step=trainer.step
        )
        if target_batch is not None:
            target_video = target_batch
            if target_video.ndim == 4:
                # TODO: This fails with 2D multi-channel targets. Handle these reliably
                target_video = target_video[:, None]
            target_video = np.swapaxes(target_video, 1, 2)
            trainer.tb.add_video(
                f'{group}_vid/target', target_video, global_step=trainer.step
            )
        for c in range(out_batch.shape[1]):
            outc_video = squash01(out_batch[:, c][None])  # Slice C, but keep dimensions intact
            # (N, C=1, T=D, H, W) -> (N, T=D, C=1, H, W)
            outc_video = np.moveaxis(outc_video, 1, 2)
            trainer.tb.add_video(
                f'{group}_vid/out{c}',
                outc_video,
                global_step=trainer.step
            )
        # TODO: Add output and target overlay videos (not straightforward
        #       because the 2D overlay code currently uses matplotlib)

    for channel in range(inp_slice.shape[0]):
        trainer.tb.add_figure(
            f'{group}/inp{channel}',
            plot_image(inp_slice[channel], cmap='gray', filename=name),
            global_step=trainer.step
        )

    if target_batch is not None:
        _out_channels = trainer.max_plot_id if is_classification else None
        _cmap = class_cmap if is_classification else continuous_cmap
        if target_slice.ndim == 2:
            trainer.tb.add_figure(
                f'{group}/target',
                plot_image(
                    target_slice, vmin=0, vmax=trainer.max_plot_id, filename=name, cmap=_cmap
                    # vmin=0., vmax=1.
                ),
                global_step=trainer.step
            )
        elif target_slice.ndim == 3:
            for c in range(target_slice.shape[0]):
                trainer.tb.add_figure(
                    f'{group}/target{c}',
                    plot_image(
                        target_slice[c], vmin=0, vmax=trainer.max_plot_id, filename=name, cmap=_cmap
                        # vmin=0., vmax=1.
                    ),
                    global_step=trainer.step
                )

    for key, img in images.items():
        if key.startswith('att'):
            trainer.tb.add_figure(
                f'{group}/{key}',
                plot_image(img, cmap='viridis'),
                global_step=trainer.step
            )

    # Plot each output channel c individually as "out{c}"
    for c in range(out_slice.shape[0]):
        trainer.tb.add_figure(
            f'{group}/out{c}',
            plot_image(
                out_slice[c], cmap=continuous_cmap, filename=name,
                # vmin=0., vmax=1.
            ),
            global_step=trainer.step
        )

    # Only make pred and overlay plots in classification scenarios
    if target_batch is not None:
        if is_classification:
            pred_slice = out_slice.argmax(0)
            trainer.tb.add_figure(
                f'{group}/pred_slice',
                plot_image(pred_slice, vmin=0, vmax=trainer.max_plot_id, cmap=class_cmap, filename=name),
                global_step=trainer.step
            )
            if target_batch is not None and not target_batch.ndim == 2:  # TODO: Make this condition more reliable and document it
                for c in range(inp_slice.shape[0]):
                    trainer.tb.add_figure(
                        f'{group}/target_overlay{c}',
                        plot_image(inp_slice[c], overlay=target_slice, overlay_alpha=trainer.overlay_alpha, vmin=0, vmax=trainer.max_plot_id, cmap=class_cmap, filename=name),
                        global_step=trainer.step
                    )
                    trainer.tb.add_figure(
                        f'{group}/pred_overlay',
                        plot_image(inp_slice[c], overlay=pred_slice, overlay_alpha=trainer.overlay_alpha, vmin=0, vmax=trainer.max_plot_id, cmap=class_cmap, filename=name),
                        global_step=trainer.step
                    )


def _tb_log_sample_images_all_img(
        trainer: 'Trainer',
        images: Dict[str, np.ndarray],
        z_plane: Optional[int] = None,
        group: str = 'sample'
) -> None:
    """Tensorboard plotting handler that plots all arrays in the ``images``
    dict as 2D grayscale images. Multi-channel images are split along the
    C dimension and plotted separately.
    """
    name = images.pop('fname', [None])[0]
    # TODO: Clean up/remove the messy name handling. Figure out how to pass non-image data cleanly.

    for key, img in images.items():
        img = img[:1]  # Always only use the first element of the batch dimension
        batch2img = _get_batch2img_function(img, z_plane)
        img = batch2img(img)
        if img.shape[0] == 1:
            trainer.tb.add_figure(
                f'{group}/{key}',
                plot_image(img[0], cmap='gray', filename=name),
                global_step=trainer.step
            )
        else:
            for c in range(img.shape[0]):
                trainer.tb.add_figure(
                    f'{group}/{key}{c}',
                    plot_image(img[c], cmap='gray', filename=name),
                    global_step=trainer.step
                )



"""Weight Standartized convolution layers, see https://arxiv.org/abs/1903.10520"""

# Adapted from https://github.com/vballoli/nfnets-pytorch/blob/61f0d6387/nfnets/base.py
# Added 3D versions of the WS layers

import torch
from torch import nn
from torch.functional import F
from torch import Tensor

from typing import Optional, List, Tuple


class FWS(nn.Module):
    """Kind of like weight standardization, but changes weights in place"""
    def __init__(
            self,
            layer: nn.Module,
            learnable_gain: bool = True,
            const_eval: bool = False,
            eps: float = 1e-4
    ):
        super().__init__()
        self.layer = layer
        self.const_eval = const_eval
        self.vmdims: Tuple[int, ...] = tuple(range(1, self.layer.weight.ndim))
        if learnable_gain:
            self.gain = nn.Parameter(
                torch.ones(self.layer.weight.shape[0], requires_grad=True)
            )
        else:
            self.register_buffer('gain', torch.ones(self.layer.weight.shape[0]))
        self.register_buffer('fan_in', torch.prod(torch.tensor(self.layer.weight.shape)))
        self.register_buffer('eps', torch.tensor(eps))

    def standardize_weight(self):  # inplace
        var, mean = torch.var_mean(self.layer.weight, dim=self.vmdims, keepdims=True)
        scale = torch.rsqrt(
            torch.max(var * self.fan_in, self.eps)
        ) * self.gain.view_as(var)
        shift = mean * scale
        with torch.no_grad():
            self.layer.weight.mul_(scale).sub_(shift)
        # return self.layer.weight * scale - shift

    def forward(self, *args, **kwargs):
        if self.training or not self.const_eval:
            self.standardize_weight()
        return self.layer(*args, **kwargs)


class WSConv3d(nn.Conv3d):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros'):
        super().__init__(in_channels, out_channels, kernel_size, stride=stride, padding=padding,
                         dilation=dilation, groups=groups, bias=bias, padding_mode=padding_mode)

        nn.init.kaiming_normal_(self.weight)
        self.gain = nn.Parameter(torch.ones(self.weight.shape[0], requires_grad=True))

    def standardize_weight(self, eps):
        var, mean = torch.var_mean(self.weight, dim=(1, 2, 3, 4), keepdims=True)
        fan_in = torch.prod(torch.tensor(self.weight.shape))

        scale = torch.rsqrt(
            torch.max(
                var * fan_in, torch.tensor(eps).to(var.device))
        ) * self.gain.view_as(var).to(var.device)
        shift = mean * scale
        return self.weight * scale - shift

    def forward(self, input, eps=1e-4):
        weight = self.standardize_weight(eps)
        return F.conv3d(input, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


class WSConvTranspose3d(nn.ConvTranspose3d):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size,
                 stride=1,
                 padding=0,
                 output_padding=0,
                 groups: int = 1,
                 bias: bool = True,
                 dilation: int = 1,
                 padding_mode: str = 'zeros'):
        super().__init__(in_channels, out_channels, kernel_size, stride=stride, padding=padding,
                         output_padding=output_padding, groups=groups, bias=bias, dilation=dilation, padding_mode=padding_mode)

        nn.init.kaiming_normal_(self.weight)
        self.gain = nn.Parameter(torch.ones(self.weight.size(0), requires_grad=True))

    def standardize_weight(self, eps):
        var, mean = torch.var_mean(self.weight, dim=(1, 2, 3, 4), keepdims=True)
        fan_in = torch.prod(torch.tensor(self.weight.shape))

        scale = torch.rsqrt(torch.max(
            var * fan_in, torch.tensor(eps).to(var.device))) * self.gain.view_as(var).to(var.device)
        shift = mean * scale
        return self.weight * scale - shift

    def forward(self, input: Tensor, output_size: Optional[List[int]] = None, eps: float = 1e-4) -> Tensor:
        weight = self.standardize_weight(eps)
        return F.conv_transpose2d(input, weight, self.bias, self.stride, self.padding, self.output_padding, self.groups, self.dilation)


class WSConv1d(nn.Conv1d):
    r"""Applies a 1D convolution over an input signal composed of several input
    planes.
    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{\text{in}}, L)` and output :math:`(N, C_{\text{out}}, L_{\text{out}})` can be
    precisely described as:
    .. math::
        \text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) +
        \sum_{k = 0}^{C_{in} - 1} \text{weight}(C_{\text{out}_j}, k)
        \star \text{input}(N_i, k)
    where :math:`\star` is the valid `cross-correlation`_ operator,
    :math:`N` is a batch size, :math:`C` denotes a number of channels,
    :math:`L` is a length of signal sequence.
    This module supports :ref:`TensorFloat32<tf32_on_ampere>`.
    * :attr:`stride` controls the stride for the cross-correlation, a single
      number or a one-element tuple.
    * :attr:`padding` controls the amount of implicit zero-paddings on both sides
      for :attr:`padding` number of points.
    * :attr:`dilation` controls the spacing between the kernel points; also
      known as the à trous algorithm. It is harder to describe, but this `link`_
      has a nice visualization of what :attr:`dilation` does.
    * :attr:`groups` controls the connections between inputs and outputs.
      :attr:`in_channels` and :attr:`out_channels` must both be divisible by
      :attr:`groups`. For example,
        * At groups=1, all inputs are convolved to all outputs.
        * At groups=2, the operation becomes equivalent to having two conv
          layers side by side, each seeing half the input channels,
          and producing half the output channels, and both subsequently
          concatenated.
        * At groups= :attr:`in_channels`, each input channel is convolved with
          its own set of filters,
          of size
          :math:`\left\lfloor\frac{out\_channels}{in\_channels}\right\rfloor`.
    Note:
        Depending of the size of your kernel, several (of the last)
        columns of the input might be lost, because it is a valid
        `cross-correlation`_, and not a full `cross-correlation`_.
        It is up to the user to add proper padding.
    Note:
        When `groups == in_channels` and `out_channels == K * in_channels`,
        where `K` is a positive integer, this operation is also termed in
        literature as depthwise convolution.
        In other words, for an input of size :math:`(N, C_{in}, L_{in})`,
        a depthwise convolution with a depthwise multiplier `K`, can be constructed by arguments
        :math:`(C_\text{in}=C_{in}, C_\text{out}=C_{in} \times K, ..., \text{groups}=C_{in})`.
    Note:
        In some circumstances when using the CUDA backend with CuDNN, this operator
        may select a nondeterministic algorithm to increase performance. If this is
        undesirable, you can try to make the operation deterministic (potentially at
        a performance cost) by setting ``torch.backends.cudnn.deterministic =
        True``.
        Please see the notes on :doc:`/notes/randomness` for background.
    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of
            the input. Default: 0
        padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``
        dilation (int or tuple, optional): Spacing between kernel
            elements. Default: 1
        groups (int, optional): Number of blocked connections from input
            channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``
    Shape:
        - Input: :math:`(N, C_{in}, L_{in})`
        - Output: :math:`(N, C_{out}, L_{out})` where
          .. math::
              L_{out} = \left\lfloor\frac{L_{in} + 2 \times \text{padding} - \text{dilation}
                        \times (\text{kernel\_size} - 1) - 1}{\text{stride}} + 1\right\rfloor
    Attributes:
        weight (Tensor): the learnable weights of the module of shape
            :math:`(\text{out\_channels},
            \frac{\text{in\_channels}}{\text{groups}}, \text{kernel\_size})`.
            The values of these weights are sampled from
            :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k = \frac{groups}{C_\text{in} * \text{kernel\_size}}`
        bias (Tensor):   the learnable bias of the module of shape
            (out_channels). If :attr:`bias` is ``True``, then the values of these weights are
            sampled from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k = \frac{groups}{C_\text{in} * \text{kernel\_size}}`
    Examples::
        >>> m = nn.Conv1d(16, 33, 3, stride=2)
        >>> input = torch.randn(20, 16, 50)
        >>> output = m(input)
    .. _cross-correlation:
        https://en.wikipedia.org/wiki/Cross-correlation
    .. _link:
        https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros'):
        super().__init__(in_channels, out_channels, kernel_size, stride=stride, padding=padding,
                         dilation=dilation, groups=groups, bias=bias, padding_mode=padding_mode)

        nn.init.kaiming_normal_(self.weight)
        self.gain = nn.Parameter(torch.ones(
            self.weight.size()[0], requires_grad=True))

    def standardize_weight(self, eps):
        var, mean = torch.var_mean(self.weight, dim=(1, 2), keepdims=True)
        fan_in = torch.prod(torch.tensor(self.weight.shape))

        scale = torch.rsqrt(torch.max(
            var * fan_in, torch.tensor(eps).to(var.device))) * self.gain.view_as(var).to(var.device)
        shift = mean * scale
        return self.weight * scale - shift

    def forward(self, input, eps=1e-4):
        weight = self.standardize_weight(eps)
        return F.conv1d(input, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


class WSConv2d(nn.Conv2d):
    """Applies a 2D convolution over an input signal composed of several input
    planes after weight normalization/standardization.
    Reference: https://github.com/deepmind/deepmind-research/blob/master/nfnets/base.py#L121
    In the simplest case, the output value of the layer with input size
    :math:`(N, C_{\text{in}}, H, W)` and output :math:`(N, C_{\text{out}}, H_{\text{out}}, W_{\text{out}})`
    can be precisely described as:
    .. math::
        \text{out}(N_i, C_{\text{out}_j}) = \text{bias}(C_{\text{out}_j}) +
        \sum_{k = 0}^{C_{\text{in}} - 1} \text{weight}(C_{\text{out}_j}, k) \star \text{input}(N_i, k)
    where :math:`\star` is the valid 2D `cross-correlation`_ operator,
    :math:`N` is a batch size, :math:`C` denotes a number of channels,
    :math:`H` is a height of input planes in pixels, and :math:`W` is
    width in pixels.
    This module supports :ref:`TensorFloat32<tf32_on_ampere>`.
    * :attr:`stride` controls the stride for the cross-correlation, a single
      number or a tuple.
    * :attr:`padding` controls the amount of implicit zero-paddings on both
      sides for :attr:`padding` number of points for each dimension.
    * :attr:`dilation` controls the spacing between the kernel points; also
      known as the à trous algorithm. It is harder to describe, but this `link`_
      has a nice visualization of what :attr:`dilation` does.
    * :attr:`groups` controls the connections between inputs and outputs.
      :attr:`in_channels` and :attr:`out_channels` must both be divisible by
      :attr:`groups`. For example,
        * At groups=1, all inputs are convolved to all outputs.
        * At groups=2, the operation becomes equivalent to having two conv
          layers side by side, each seeing half the input channels,
          and producing half the output channels, and both subsequently
          concatenated.
        * At groups= :attr:`in_channels`, each input channel is convolved with
          its own set of filters, of size:
          :math:`\left\lfloor\frac{out\_channels}{in\_channels}\right\rfloor`.
    The parameters :attr:`kernel_size`, :attr:`stride`, :attr:`padding`, :attr:`dilation` can either be:
        - a single ``int`` -- in which case the same value is used for the height and width dimension
        - a ``tuple`` of two ints -- in which case, the first `int` is used for the height dimension,
          and the second `int` for the width dimension
    Note:
         Depending of the size of your kernel, several (of the last)
         columns of the input might be lost, because it is a valid `cross-correlation`_,
         and not a full `cross-correlation`_.
         It is up to the user to add proper padding.
    Note:
        When `groups == in_channels` and `out_channels == K * in_channels`,
        where `K` is a positive integer, this operation is also termed in
        literature as depthwise convolution.
        In other words, for an input of size :math:`(N, C_{in}, H_{in}, W_{in})`,
        a depthwise convolution with a depthwise multiplier `K`, can be constructed by arguments
        :math:`(in\_channels=C_{in}, out\_channels=C_{in} \times K, ..., groups=C_{in})`.
    Note:
        In some circumstances when using the CUDA backend with CuDNN, this operator
        may select a nondeterministic algorithm to increase performance. If this is
        undesirable, you can try to make the operation deterministic (potentially at
        a performance cost) by setting ``torch.backends.cudnn.deterministic =
        True``.
        Please see the notes on :doc:`/notes/randomness` for background.
    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of
            the input. Default: 0
        padding_mode (string, optional): ``'zeros'``, ``'reflect'``,
            ``'replicate'`` or ``'circular'``. Default: ``'zeros'``
        dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
        groups (int, optional): Number of blocked connections from input
            channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the
            output. Default: ``True``
    Shape:
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where
          .. math::
              H_{out} = \left\lfloor\frac{H_{in}  + 2 \times \text{padding}[0] - \text{dilation}[0]
                        \times (\text{kernel\_size}[0] - 1) - 1}{\text{stride}[0]} + 1\right\rfloor
          .. math::
              W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] - \text{dilation}[1]
                        \times (\text{kernel\_size}[1] - 1) - 1}{\text{stride}[1]} + 1\right\rfloor
    Attributes:
        weight (Tensor): the learnable weights of the module of shape
            :math:`(\text{out\_channels}, \frac{\text{in\_channels}}{\text{groups}},`
            :math:`\text{kernel\_size[0]}, \text{kernel\_size[1]})`.
            The values of these weights are sampled from
            :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k = \frac{groups}{C_\text{in} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
        bias (Tensor):   the learnable bias of the module of shape
            (out_channels). If :attr:`bias` is ``True``,
            then the values of these weights are
            sampled from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
            :math:`k = \frac{groups}{C_\text{in} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
    Examples:
        >>> # With square kernels and equal stride
        >>> m = WSConv2d(16, 33, 3, stride=2)
        >>> # non-square kernels and unequal stride and with padding
        >>> m = WSConv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2))
        >>> # non-square kernels and unequal stride and with padding and dilation
        >>> m = WSConv2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2), dilation=(3, 1))
        >>> input = torch.randn(20, 16, 50, 100)
        >>> output = m(input)
    .. _cross-correlation:
        https://en.wikipedia.org/wiki/Cross-correlation
    .. _link:
        https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros'):
        super().__init__(in_channels, out_channels, kernel_size, stride=stride, padding=padding,
                         dilation=dilation, groups=groups, bias=bias, padding_mode=padding_mode)

        nn.init.kaiming_normal_(self.weight)
        self.gain = nn.Parameter(torch.ones(
            self.weight.size(0), requires_grad=True))

    def standardize_weight(self, eps):
        var, mean = torch.var_mean(self.weight, dim=(1, 2, 3), keepdims=True)
        fan_in = torch.prod(torch.tensor(self.weight.shape[0:]))

        scale = torch.rsqrt(torch.max(
            var * fan_in, torch.tensor(eps).to(var.device))) * self.gain.view_as(var).to(var.device)
        shift = mean * scale
        return self.weight * scale - shift

    def forward(self, input, eps=1e-4):
        weight = self.standardize_weight(eps)
        return F.conv2d(input, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


class WSConvTranspose2d(nn.ConvTranspose2d):
    """Applies a 2D transposed convolution operator over an input image
    composed of several input planes after weight normalization/standardization.
    This module can be seen as the gradient of Conv2d with respect to its input.
    It is also known as a fractionally-strided convolution or
    a deconvolution (although it is not an actual deconvolution operation).
    This module supports :ref:`TensorFloat32<tf32_on_ampere>`.
    * :attr:`stride` controls the stride for the cross-correlation.
    * :attr:`padding` controls the amount of implicit zero-paddings on both
      sides for ``dilation * (kernel_size - 1) - padding`` number of points. See note
      below for details.
    * :attr:`output_padding` controls the additional size added to one side
      of the output shape. See note below for details.
    * :attr:`dilation` controls the spacing between the kernel points; also known as the à trous algorithm.
      It is harder to describe, but this `link`_ has a nice visualization of what :attr:`dilation` does.
    * :attr:`groups` controls the connections between inputs and outputs.
      :attr:`in_channels` and :attr:`out_channels` must both be divisible by
      :attr:`groups`. For example,
        * At groups=1, all inputs are convolved to all outputs.
        * At groups=2, the operation becomes equivalent to having two conv
          layers side by side, each seeing half the input channels,
          and producing half the output channels, and both subsequently
          concatenated.
        * At groups= :attr:`in_channels`, each input channel is convolved with
          its own set of filters (of size
          :math:`\left\lfloor\frac{out\_channels}{in\_channels}\right\rfloor`).
    The parameters :attr:`kernel_size`, :attr:`stride`, :attr:`padding`, :attr:`output_padding`
    can either be:
        - a single ``int`` -- in which case the same value is used for the height and width dimensions
        - a ``tuple`` of two ints -- in which case, the first `int` is used for the height dimension,
          and the second `int` for the width dimension
    .. note::
         Depending of the size of your kernel, several (of the last)
         columns of the input might be lost, because it is a valid `cross-correlation`_,
         and not a full `cross-correlation`_.
         It is up to the user to add proper padding.
    Note:
        The :attr:`padding` argument effectively adds ``dilation * (kernel_size - 1) - padding``
        amount of zero padding to both sizes of the input. This is set so that
        when a :class:`~torch.nn.Conv2d` and a :class:`~torch.nn.ConvTranspose2d`
        are initialized with same parameters, they are inverses of each other in
        regard to the input and output shapes. However, when ``stride > 1``,
        :class:`~torch.nn.Conv2d` maps multiple input shapes to the same output
        shape. :attr:`output_padding` is provided to resolve this ambiguity by
        effectively increasing the calculated output shape on one side. Note
        that :attr:`output_padding` is only used to find output shape, but does
        not actually add zero-padding to output.
    Note:
        In some circumstances when using the CUDA backend with CuDNN, this operator
        may select a nondeterministic algorithm to increase performance. If this is
        undesirable, you can try to make the operation deterministic (potentially at
        a performance cost) by setting ``torch.backends.cudnn.deterministic =
        True``.
        Please see the notes on :doc:`/notes/randomness` for background.
    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the convolution
        kernel_size (int or tuple): Size of the convolving kernel
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): ``dilation * (kernel_size - 1) - padding`` zero-padding
            will be added to both sides of each dimension in the input. Default: 0
        output_padding (int or tuple, optional): Additional size added to one side
            of each dimension in the output shape. Default: 0
        groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
        bias (bool, optional): If ``True``, adds a learnable bias to the output. Default: ``True``
        dilation (int or tuple, optional): Spacing between kernel elements. Default: 1
    Shape:
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where
        .. math::
              H_{out} = (H_{in} - 1) \times \text{stride}[0] - 2 \times \text{padding}[0] + \text{dilation}[0]
                        \times (\text{kernel\_size}[0] - 1) + \text{output\_padding}[0] + 1
        .. math::
              W_{out} = (W_{in} - 1) \times \text{stride}[1] - 2 \times \text{padding}[1] + \text{dilation}[1]
                        \times (\text{kernel\_size}[1] - 1) + \text{output\_padding}[1] + 1
    Attributes:
        weight (Tensor): the learnable weights of the module of shape
                         :math:`(\text{in\_channels}, \frac{\text{out\_channels}}{\text{groups}},`
                         :math:`\text{kernel\_size[0]}, \text{kernel\_size[1]})`.
                         The values of these weights are sampled from
                         :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                         :math:`k = \frac{groups}{C_\text{out} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
        bias (Tensor):   the learnable bias of the module of shape (out_channels)
                         If :attr:`bias` is ``True``, then the values of these weights are
                         sampled from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                         :math:`k = \frac{groups}{C_\text{out} * \prod_{i=0}^{1}\text{kernel\_size}[i]}`
    Examples::
        >>> # With square kernels and equal stride
        >>> m = WSConvTranspose2d(16, 33, 3, stride=2)
        >>> # non-square kernels and unequal stride and with padding
        >>> m = WSConvTranspose2d(16, 33, (3, 5), stride=(2, 1), padding=(4, 2))
        >>> input = torch.randn(20, 16, 50, 100)
        >>> output = m(input)
        >>> # exact output size can be also specified as an argument
        >>> input = torch.randn(1, 16, 12, 12)
        >>> downsample = WSConv2d(16, 16, 3, stride=2, padding=1)
        >>> upsample = WSConvTranspose2d(16, 16, 3, stride=2, padding=1)
        >>> h = downsample(input)
        >>> h.size()
        torch.Size([1, 16, 6, 6])
        >>> output = upsample(h, output_size=input.size())
        >>> output.size()
        torch.Size([1, 16, 12, 12])
    .. _cross-correlation:
        https://en.wikipedia.org/wiki/Cross-correlation
    .. _link:
        https://github.com/vdumoulin/conv_arithmetic/blob/master/README.md
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size,
                 stride=1,
                 padding=0,
                 output_padding=0,
                 groups: int = 1,
                 bias: bool = True,
                 dilation: int = 1,
                 padding_mode: str = 'zeros'):
        super().__init__(in_channels, out_channels, kernel_size, stride=stride, padding=padding,
                         output_padding=output_padding, groups=groups, bias=bias, dilation=dilation, padding_mode=padding_mode)

        nn.init.kaiming_normal_(self.weight)
        self.gain = nn.Parameter(torch.ones(
            self.weight.size(0), requires_grad=True))

    def standardize_weight(self, eps):
        var, mean = torch.var_mean(self.weight, dim=(1, 2, 3), keepdims=True)
        fan_in = torch.prod(torch.tensor(self.weight.shape[0:]))

        scale = torch.rsqrt(torch.max(
            var * fan_in, torch.tensor(eps).to(var.device))) * self.gain.view_as(var).to(var.device)
        shift = mean * scale
        return self.weight * scale - shift

    def forward(self, input: Tensor, output_size: Optional[List[int]] = None, eps: float = 1e-4) -> Tensor:
        weight = self.standardize_weight(eps)
        return F.conv_transpose2d(input, self.weight, self.bias, self.stride, self.padding, self.output_padding, self.groups, self.dilation)


# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch and others (see individual docstrings for more author info)

"""Utilities for neural network model manipulation"""

from typing import Tuple, Union, List

import numpy as np
import torch
from torch import nn


def visualize_receptive_field(
        model: torch.nn.Module,
        input_shape: Tuple[int, ...] = (96, 96, 1),  # (H, W, C) order!
        interactive=False
) -> None:
    """Visualize analytical and effective receptive filelds of a network.

    Requires https://github.com/fornaxai/receptivefield to work
    (``pip install receptivefield``).

    Limitations:
    - Only works for 2D inputs (tensor layout (N, C, H, W)!
    - Doesn't work with ScriptModules (yet). As a workaround, you can
      construct a new model (``pymodel``) with the same code (but without
      compiling it) and set its ``state_dict`` to the ``state_dict`` of the
      compiled model (``scriptmodel``) as follows:
      >>> # Python-defined standard PyTorch model
      >>> pymodel: torch.nn.Module
      >>> # Equivalent TorchScript representation of pymodel
      >>> scriptmodel: torch.jit.ScriptModule
      >>> # Update pymodel state
      >>> pymodel.load_state_dict(scriptmodel.state_dict())
      Then you can safely pass ``pymodel`` to ``visualize_receptive_field``.
    - Requires that the ``model`` stores its output (or an intermediate
      layer output of interest) as ``self.feature_maps[0]`` before returning,
      (see bottom of :py:meth:`elektronn3.models.unet.UNet.forward`).

    Example::
    >>> from elektronn3.models._model_utils import visualize_receptive_field
    >>> from elektronn3.models.unet import UNet
    >>> model = UNet(
    ...     in_channels=1,
    ...     out_channels=2,
    ...     n_blocks=3,
    ...     activation='lin',
    ...     dim=2,
    ... )
    >>> # (Train model for a few iterations here)
    >>> input_shape = (96, 96, 1)
    >>> visualize_receptive_field(model, input_shape, interactive=True)

    input_shape = (96, 96, 1)
    visualize_receptive_field(model, input_shape, interactive=True)"""
    from receptivefield.pytorch import PytorchReceptiveField
    import matplotlib.pyplot as plt

    def model_fn():
        return model.eval()

    rf = PytorchReceptiveField(model_fn)
    rf_params = rf.compute(input_shape=input_shape)
    center = (input_shape[0] // 2, input_shape[1] // 2)

    fig, ax = plt.subplots()
    rf.plot_gradient_at(fm_id=0, point=center, axis=ax)
    if interactive:
        plt.show()
    return fig


def find_first(model: nn.Module, module_type: type = nn.Conv2d) -> Tuple[str, nn.Module]:
    """Return the first submodule of an nn.Module with a certain type."""
    for name, mod in model.named_modules():
        if isinstance(mod, module_type):
            return name, mod
    # Loop finished, nothing found...
    raise LookupError(f'Module does\'nt have any layers of type {module_type}.')


def find_first_conv(model: nn.Module, in_channels=3) -> Tuple[str, nn.Module]:
    """Return the first Conv2d submodule with a certain in_channels value."""
    for name, mod in model.named_modules():
        if isinstance(mod, nn.Conv2d) and mod.in_channels == in_channels:
            return name, mod
    # Loop finished, nothing found...
    raise LookupError(f'Module does\'nt have any Conv2d layers with in_channels={in_channels}.')


# TODO: Layer initialization?
# TODO: Is it possible to re-use original conv1 weights in some way?
def change_conv1_input_channels(
        model: nn.Module,
        old_in_channels: int = 3,
        new_in_channels: int = 1
) -> None:
    """Change input channels of a convnet model (in-place).

    This can be used to turn RGB models (``in_channels=3``) into single-channel
    models (``in_channels=1``) automatically by replacing the first Conv layer,
    while preserving pretrained weights in all other layers.."""
    name, conv1 = find_first_conv(model, in_channels=old_in_channels)
    model.__dict__[name] = nn.Conv2d(
        new_in_channels, conv1.out_channels, conv1.kernel_size,
        conv1.stride, conv1.padding, conv1.dilation, conv1.groups, conv1.bias
    )


def num_params(model: torch.nn.Module) -> int:
    """Total number of trainable parameters."""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


# TODO: Support passing example tensors directly instead of shapes
def model_summary(
        model: nn.Module,
        in_shapes: Union[Tuple[int, ...], List[Tuple[int, ...]]],
        device: Union[torch.device, str] = 'cpu'
) -> Tuple[dict, str]:
    """Get a summ_dict of what a model does

    Based on https://github.com/sksq96/pytorch-summ_dict/blob/b50f21/torchsummary/torchsummary.py
    by Shubham Chandel and others
    (see contributors listed in https://github.com/sksq96/pytorch-summary).

    but  with many changes such as
    - torch.device support
    - Cleaner code using PyTorch 1.0 API and requiring Python 3.6
    - Reporting of final output shape
    - Not printing results directly, but returning them as a string and a dict
    - No more special treatment of the batch dimension
    """

    def register_hook(module):

        def hook(module, input, output):
            class_name = str(module.__class__).split('.')[-1].split("'")[0]
            module_idx = len(summ_dict)

            m_key = '%s-%i' % (class_name, module_idx + 1)
            summ_dict[m_key] = {}  # OrderedDict()
            summ_dict[m_key]['input_shape'] = list(input[0].shape)
            if isinstance(output, (list, tuple)):
                summ_dict[m_key]['output_shape'] = [
                    list(o.shape) for o in output
                ]
            else:
                summ_dict[m_key]['output_shape'] = list(output.shape)

            params = 0
            if hasattr(module, 'weight') and hasattr(module.weight, 'shape'):
                params += module.weight.numel()
                summ_dict[m_key]['trainable'] = module.weight.requires_grad
            if hasattr(module, 'bias') and hasattr(module.bias, 'shape'):
                params += module.bias.numel()
            summ_dict[m_key]['nb_params'] = params

        if (
            not isinstance(module, nn.Sequential)
            and not isinstance(module, nn.ModuleList)
            and not (module == model)
        ):
            hooks.append(module.register_forward_hook(hook))

    # multiple inputs to the network
    if isinstance(in_shapes, tuple):
        in_shapes = [in_shapes]

    x = [torch.randn(in_shape, device=device) for in_shape in in_shapes]

    summ_dict = {}  # OrderedDict()
    hooks = []

    model.apply(register_hook)

    # make a forward pass
    model(*x)

    # remove these hooks
    for h in hooks:
        h.remove()

    summ_list = []  # For building the string representation

    summ_list.append('----------------------------------------------------------------------------')
    line_new = '{:>25}  {:>28} {:>20}'.format('Layer (type)', 'Output Shape', 'Param #')
    summ_list.append(line_new)
    summ_list.append('============================================================================')
    total_params = 0
    total_output = 0
    trainable_params = 0
    for layer in summ_dict:
        # input_shape, output_shape, trainable, nb_params
        line_new = '{:>28}  {:>28} {:>15}'.format(
            layer,
            str(summ_dict[layer]['output_shape']),
            '{0:,}'.format(summ_dict[layer]['nb_params']),
        )
        total_params += summ_dict[layer]['nb_params']
        total_output += np.prod(summ_dict[layer]['output_shape'])
        if 'trainable' in summ_dict[layer]:
            if summ_dict[layer]['trainable']:
                trainable_params += summ_dict[layer]['nb_params']
        summ_list.append(line_new)

    # assume 4 bytes/number (float on cuda).
    total_in_shape = abs(np.prod(in_shapes) * 4. / (1024 ** 2.))
    total_output_size = abs(2. * total_output * 4. / (1024 ** 2.))  # x2 for gradients
    total_params_size = abs(total_params * 4. / (1024 ** 2.))
    total_size = total_params_size + total_output_size + total_in_shape
    last_key = list(summ_dict.keys())[-1]
    output_shape = summ_dict[last_key]['output_shape']

    summ_dict['total_in_shape'] = total_in_shape
    summ_dict['total_output_size'] = total_output_size
    summ_dict['total_params_size'] = total_params_size
    summ_dict['total_size'] = total_size
    summ_dict['output_shape'] = output_shape

    summ_list.append('============================================================================')
    summ_list.append('Total params: {0:,}'.format(total_params))
    summ_list.append('Trainable params: {0:,}'.format(trainable_params))
    summ_list.append('Non-trainable params: {0:,}'.format(total_params - trainable_params))
    summ_list.append('----------------------------------------------------------------------------')
    summ_list.append('Input size (MB): %0.2f' % total_in_shape)
    summ_list.append('Forward/backward pass size (MB): %0.2f' % total_output_size)
    summ_list.append('Params size (MB): %0.2f' % total_params_size)
    summ_list.append('Estimated Total Size (MB): %0.2f' % total_size)
    summ_list.append('Output shape: {}'.format(output_shape))
    summ_list.append('----------------------------------------------------------------------------')

    summ_str = '\n'.join(summ_list)

    return summ_dict, summ_str



# -*- coding: utf-8 -*-
# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Marius Killinger, Martin Drawitsch, Philipp Schubert

import logging
import os

import numpy as np
from matplotlib import pyplot as plt
logger = logging.getLogger('elektronn3log')


class Scroller:
    def __init__(self, axes, images, names, init_z=None):
        self.axes = axes
        for ax in axes:
            ax.grid(b=False)
        # ax.set_title('use scroll wheel to navigate images')

        self.images = list(map(np.ascontiguousarray, images))
        self.n_slices = images[0].shape[0]
        self.ind = self.n_slices // 2 if init_z is None else init_z

        self.imgs = []
        for ax, dat, name in zip(axes, images, names):
            if name in ['id', 'ids', 'ID', 'IDs', 'seg', 'SEG', 'Seg', 'lab',
                        'label', 'Label']:
                cmap = 'nipy_spectral'
            else:
                cmap = 'gray'

            self.imgs.append(
                ax.imshow(dat[self.ind], interpolation='None', cmap=cmap))
            ax.set_xlabel(name)

        self.update()

    def onscroll(self, event):
        # print ("%s %s" % (event.button, event.step))
        if event.button=='up':
            self.ind = np.clip(self.ind + 1, 0, self.n_slices - 1)
        else:
            self.ind = np.clip(self.ind - 1, 0, self.n_slices - 1)

        self.update()

    def update(self):
        for ax, im, dat in zip(self.axes, self.imgs, self.images):
            im.set_data(dat[self.ind])
            ax.set_ylabel('slice %s' % self.ind)
            im.axes.figure.canvas.draw()


def scroll_plot(images, names=None, init_z=None):
    """
    Creates a plot 1x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of arrays (or single)
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of strings (or single)
      Names for each image

    Usage
    -----

    For the scroll interaction to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    if names is None:
        n = 1 if isinstance(images, np.ndarray) else len(images)
        names = [str(i) for i in range(n)]

    if isinstance(images, np.ndarray):
        return _scroll_plot1(images, names, init_z)
    elif len(images)==2:
        assert len(names) >= 2
        return _scroll_plot2(images, names, init_z)
    elif len(images)==4:
        assert len(names) >= 4
        return _scroll_plot4(images, names, init_z)


def _scroll_plot1(image, name, init_z):
    """
    Creates a plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images:  array of shape (z,x,y) or (z,x,y,RGB)

    Usage
    -----

    For the scroll interaction to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = scroll_plot(image, name)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(111)

    scroller = Scroller([ax1], [image, ], [name, ], init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def _scroll_plot2(images, names, init_z):
    """
    Creates a plot 1x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of 2 arrays
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of 2 strings
      Names for each image

    Usage
    -----

    For the scroll interaction to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 6))
    ax1 = fig.add_subplot(121)
    ax2 = fig.add_subplot(122, sharex=ax1, sharey=ax1)

    scroller = Scroller([ax1, ax2], images, names, init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def _scroll_plot4(images, names, init_z):
    """
    Creates a plot 2x2 image plot of 3d volume images
    Scrolling changes the displayed slices

    Parameters
    ----------

    images: list of 4 arrays
      Each array of shape (z,y,x) or (z,y,x,RGB)
    names: list of 4 strings
      Names for each image

    Usage
    -----

    For the scroll interaction to work, the "scroller" object
    must be returned to the calling scope

    >>> fig, scroller = _scroll_plot4(images, names)
    >>> fig.show()

    """
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(221)
    ax2 = fig.add_subplot(222, sharex=ax1, sharey=ax1)
    ax3 = fig.add_subplot(223, sharex=ax1, sharey=ax1)
    ax4 = fig.add_subplot(224, sharex=ax1, sharey=ax1)

    scroller = Scroller([ax1, ax2, ax3, ax4], images, names, init_z)
    fig.canvas.mpl_connect('scroll_event', scroller.onscroll)
    fig.tight_layout()
    return scroller


def sma(c, n):
    """
    Returns box-SMA of c with box length n, the returned array has the same
    length as c and is const-padded at the beginning
    """
    if n==0:
        return c
    ret = np.cumsum(c, dtype=float)
    ret[n:] = (ret[n:] - ret[:-n]) / n
    m = min(n, len(c))
    ret[:n] = ret[:n] / np.arange(1, m + 1)  # unsmoothed
    return ret


def add_timeticks(ax, times, steps, time_str='mins', num=5):
    N = int(times[-1])
    k = max(N / num, 1)
    k = int(np.log10(k))  # 10-base of locators
    m = int(np.round(float(N) / (num * 10 ** k)))  # multiple of base
    s = max(m * 10 ** k, 1)
    x_labs = np.arange(0, N, s, dtype=np.int64)
    x_ticks = np.interp(x_labs, times, steps)
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(x_labs)
    ax.set_xlim(0, steps[-1])
    ax.set_xlabel('Runtime [%s]' % time_str)  # (%s)'%("{0:,d}".format(N)))


def plot_hist(timeline, history, save_path, loss_smoothing_length=200,
              autoscale=True):
    """Plot graphical info during Training"""
    plt.ioff()
    try:
        # Subsample points for plotting
        N = len(timeline)
        x_timeline = np.arange(N)
        s = max((len(timeline) // 2000), 1)
        x_timeline = x_timeline[::s]
        timeline = timeline[::s]
        s = max((len(history) // 2000), 1)
        history = history[::s]

        if timeline['time'][-1] < 120 * 60:
            runtime = str(int(timeline['time'][-1] / 60)) + ' mins'
        else:
            runtime = "%.1f hours" % (timeline['time'][-1] / 3600)

        # check if valid data is available
        if not np.any(np.isnan(history['valid_loss'])):
            l = history['valid_loss'][-10:]
        else:
            l = history['train_loss'][-10:]

        loss_cap = l.mean() + 2 * l.std()
        lt = timeline['loss'][-200:]
        lt_m = lt.mean()
        lt_s = lt.std()
        loss_cap_t = lt_m + 2 * lt_s
        loss_cap = np.maximum(loss_cap, loss_cap_t)

        if np.all(timeline['loss'] > 0):
            loss_floor = 0.0
        else:
            loss_floor = lt_m - 2 * lt_s

        ### Timeline, Loss ###
        plt.figure(figsize=(16, 12))
        plt.subplot(211)
        plt.plot(x_timeline, timeline['loss'], 'b-', alpha=0.5,
                 label='Update Loss')
        loss_smooth = sma(timeline['loss'], loss_smoothing_length)
        plt.plot(x_timeline, loss_smooth, 'k-', label='Smooth update Loss',
                 linewidth=3)

        if autoscale:
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(0, N)

        plt.legend(loc=0)
        plt.hlines(lt_m, 0, N, linestyle='dashed', colors='r', linewidth=2)
        plt.hlines(lt_m + lt_s, 0, N, linestyle='dotted', colors='r',
                   linewidth=1)
        plt.hlines(lt_m - lt_s, 0, N, linestyle='dotted', colors='r',
                   linewidth=1)
        plt.xlabel('Update steps %s, total runtime %s' % (N - 1, runtime))

        ax = plt.twiny()
        if timeline['time'][-1] > 120 * 60:
            add_timeticks(ax, timeline['time'] / 3600, x_timeline,
                          time_str='hours')
        else:
            add_timeticks(ax, timeline['time'] / 60, x_timeline,
                          time_str='mins')

        ### Loss vs Prevalence ###
        plt.subplot(212)
        c = 1.0 - (timeline['time'] / timeline['time'].max())
        plt.scatter(timeline['batch_char'], timeline['loss'], c=c, marker='.',
                    s=80, cmap='gray', edgecolors='face')
        if autoscale:
            bc = timeline['batch_char'][-200:]
            bc_m = bc.mean()
            bc_s = bc.std()
            bc_cap = bc_m + 2 * bc_s
            if np.all(bc > 0):
                bc_floor = -0.01
            else:
                bc_floor = bc_m - 2 * bc_s
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(bc_floor, bc_cap + 1e-10)

        plt.xlabel('Mean target of batch')
        plt.ylabel('Loss')

        plt.tight_layout()
        plt.savefig(os.path.join(save_path, 'timeline.png'), bbox_inches='tight')

        ###################################################################
        ### History Loss ###
        plt.figure(figsize=(16, 12))
        plt.subplot(311)
        plt.plot(history['steps'], history['train_loss'], 'g-',
                 label='Train Loss', linewidth=3)
        plt.plot(history['steps'], history['valid_loss'], 'r-',
                 label='Valid Loss', linewidth=3)
        if autoscale:
            plt.ylim(loss_floor, loss_cap)
            plt.xlim(0, history['steps'][-1])

        plt.legend(loc=0)
        # plt.xlabel('Update steps %s, total runtime %s'%(N-1, runtime))

        ax = plt.twiny()
        if timeline['time'][-1] > 120 * 60:
            add_timeticks(ax, timeline['time'] / 3600, x_timeline,
                          time_str='hours')
        else:
            add_timeticks(ax, timeline['time'] / 60, x_timeline,
                          time_str='mins')

        ### History Loss gains ###
        plt.subplot(312)
        plt.plot(history['steps'], history['loss_gain'], 'b-',
                 label='Loss Gain at update', linewidth=3)
        plt.hlines(0, 0, history['steps'][-1], linestyles='dotted')
        plt.plot(history['steps'], history['lr'], 'r-', label='LR',
                 linewidth=3)
        # plt.xlabel('Update steps %s, total runtime %s'%(N-1, runtime))
        plt.legend(loc=3)
        std = history['loss_gain'][:5].std() * 2 if len(history) > 6 else 1.0
        if autoscale:
            # add epsilon to suppress matplotlib warning in case of CG
            plt.ylim(-std, std + 1e-10)
            plt.xlim(0, history['steps'][-1])

        ax2 = plt.twinx()
        ax2.plot(history['steps'], history['mom'], 'r-', label='MOM')
        ax2.plot(history['steps'], history['gradnetrate'], 'r-',
                 label='GradNetRate')

        ax2.set_ylim(-1, 1)
        if autoscale:
            ax2.set_xlim(0, history['steps'][-1])
        ax2.legend(loc=4)

        ### Errors ###
        plt.subplot(313)
        cutoff = 2
        if len(history) > (cutoff + 1):
            history = history[cutoff:]
        # check if valid data is available
        if not np.any(np.isnan(history['valid_err'])):
            e = history['valid_err'][-200:]
        else:
            e = history['train_err'][-200:]
        e_m = e.mean()
        e_s = e.std()
        err_cap = e_m + 2 * e_s
        if np.all(e > 0):
            err_floor = 0.0
        else:
            err_floor = e_m - 2 * e_s

        plt.plot(history['steps'], history['train_err'], 'g--',
                 label='Train error', linewidth=1)
        plt.plot(history['steps'], history['valid_err'], 'r--',
                 label='Valid Error', linewidth=1)

        plt.plot(history['steps'], sma(history['train_err'], 8), 'g-',
                 label='Smooth train error', linewidth=3)
        if not np.any(np.isnan(sma(history['valid_err'], 8))):
            plt.plot(history['steps'], sma(history['valid_err'], 8), 'r-',
                     label='Smooth valid Error', linewidth=3)
        if autoscale:
            try:
                plt.ylim(err_floor, err_cap)
            except ValueError:  # NaN errors can't be used for auto-scaling.
                pass
            plt.xlim(0, history['steps'][-1])

        plt.grid()
        plt.legend(loc=0)
        plt.xlabel('Update steps %s, total runtime %s' % (N - 1, runtime))
        plt.tight_layout()
        plt.savefig(os.path.join(save_path, 'history.png'), bbox_inches='tight')

    except ValueError:
        # When arrays are empty
        logger.exception("An error occurred during plotting.")



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch

"""Loss functions"""
from typing import Sequence, Optional, Tuple, Callable, Union

import torch
import numpy as np

from torch import nn
from torch.nn import functional as F

from elektronn3.modules.lovasz_losses import lovasz_softmax


class CombinedLoss(torch.nn.Module):
    """Defines a loss function as a weighted sum of combinable loss criteria.

    Args:
        criteria: List of loss criterion modules that should be combined.
        weight: Weight assigned to the individual loss criteria (in the same
            order as ``criteria``).
        device: The device on which the loss should be computed. This needs
            to be set to the device that the loss arguments are allocated on.
    """
    def __init__(
            self,
            criteria: Sequence[torch.nn.Module],
            weight: Optional[Sequence[float]] = None,
            device: torch.device = None
    ):
        super().__init__()
        self.criteria = torch.nn.ModuleList(criteria)
        self.device = device
        if weight is None:
            weight = torch.ones(len(criteria))
        else:
            weight = torch.as_tensor(weight, dtype=torch.float32)
            assert weight.shape == (len(criteria),)
        self.register_buffer('weight', weight.to(self.device))

    def forward(self, *args):
        loss = torch.tensor(0., device=self.device)
        for crit, weight in zip(self.criteria, self.weight):
            loss += weight * crit(*args)
        return loss


class FocalLoss(torch.nn.Module):
    """Focal Loss (https://arxiv.org/abs/1708.02002)
    
    Expects raw outputs, not softmax probs."""
    def __init__(self, weight=None, gamma=2., reduction='mean', ignore_index=-100):
        super().__init__()
        self.gamma = gamma
        self.nll = torch.nn.NLLLoss(weight=weight, reduction=reduction, ignore_index=ignore_index)
        self.log_softmax = torch.nn.LogSoftmax(1)
        
    def forward(self, output, target):
        log_prob = self.log_softmax(output)
        prob = torch.exp(log_prob)
        return self.nll(((1 - prob) ** self.gamma) * log_prob, target)


class SoftmaxBCELoss(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.bce = torch.nn.BCELoss(*args, **kwargs)

    def forward(self, output, target):
        probs = torch.nn.functional.softmax(output, dim=1)
        return self.bce(probs, target)

# class MultiLabelCrossEntropy(nn.Module):
#     def __init__(self, weight=torch.tensor(1.)):
#         self.register_buffer('weight', weight)

#     def forward(self, output, target):
#         assert output.shape == target.shape
#         logprobs = torch.nn.functional.log_softmax(output, dim=1)
#         wsum = self.weight[None] * torch.sum(-target * logprobs)
#         return torch.mean(wsum, dim=1)


def global_average_pooling(inp: torch.Tensor) -> torch.Tensor:
    if inp.ndim == 5:
        return F.adaptive_avg_pool3d(inp, 1)
    elif inp.ndim == 4:
        return F.adaptive_avg_pool2d(inp, 1)
    else:
        raise NotImplementedError


class GAPTripletMarginLoss(nn.TripletMarginLoss):
    """Same as ``torch.nn.TripletMarginLoss``, but applies global average
    pooling to anchor, positive and negative tensors before calculating the
    loss."""

    def forward(self, anchor: torch.Tensor, positive: torch.Tensor, negative: torch.Tensor) -> torch.Tensor:
        return super().forward(
            global_average_pooling(anchor),
            global_average_pooling(positive),
            global_average_pooling(negative)
        )


class MaskedMSELoss(nn.Module):
    """Masked MSE loss where only pixels that are masked are considered.

    Expects an optional binary mask as the third argument.
    If no mask is supplied (``None``), the loss is equivalent to ``torch.nn.MSELoss``."""
    @staticmethod
    def forward(out, target, mask=None):
        if mask is None:
            return F.mse_loss(out, target)
        err = F.mse_loss(out, target, reduction='none')
        err *= mask
        loss = err.sum() / mask.sum()  # Scale by ratio of masked pixels
        return loss


class DistanceWeightedMSELoss(nn.Module):
    """Weighted MSE loss for signed euclidean distance transform targets.

    By setting ``fg_weight`` to a high value, the errors in foreground
    regions are more strongly penalized.
    If ``fg_weight=1``, this loss is equivalent to ``torch.nn.MSELoss``.

    Requires that targets are transformed with
    :py:class:`elektronn3.data.transforms.DistanceTransformTarget`

    Per-pixel weights are assigned on the targets as follows:
    - each foreground pixel is weighted by ``fg_weight``
    - each background pixel is weighted by 1.
    """
    def __init__(self, fg_weight=100., mask_borders=40):
        super().__init__()
        self.fg_weight = fg_weight
        self.mask_borders = mask_borders

    def forward(self, output, target):
        mse = nn.functional.mse_loss(output, target, reduction='none')
        with torch.no_grad():
            # This assumes that the target is in the (-1, 1) value range (tanh)
            #  and that regions with value <= 0 are foreground.
            weight = torch.ones_like(target)
            weight[target <= 0] = self.fg_weight
            if self.mask_borders is not None:  # Mask out invalid regions that come from same-padding
                o = self.mask_borders
                weight[:, :, :o, :o] = 0.
                weight[:, :, target.shape[-2] - o:, target.shape[-1] - o:] = 0.
        return torch.mean(weight * mse)


def _channelwise_sum(x: torch.Tensor):
    """Sum-reduce all dimensions of a tensor except dimension 1 (C)"""
    reduce_dims = tuple([0] + list(range(x.dim()))[2:])  # = (0, 2, 3, ...)
    return x.sum(dim=reduce_dims)


# TODO: Dense weight support
def dice_loss(probs, target, weight=1., eps=0.0001, smooth=0.):
    # Probs need to be softmax probabilities, not raw network outputs
    tsh, psh = target.shape, probs.shape

    if tsh == psh:  # Already one-hot
        onehot_target = target.to(probs.dtype)
    elif tsh[0] == psh[0] and tsh[1:] == psh[2:]:  # Assume dense target storage, convert to one-hot
        onehot_target = torch.zeros_like(probs)
        onehot_target.scatter_(1, target.unsqueeze(1), 1)
    else:
        raise ValueError(
            f'Target shape {target.shape} is not compatible with output shape {probs.shape}.'
        )
    # if weight is None:
    #     weight = torch.ones(probs.shape[0], dtype=probs.dtype)  # (C,)
    # if ignore_index is not None:
    #     weight[:, ignore_index] = 0.

    intersection = probs * onehot_target  # (N, C, ...)
    numerator = 2 * _channelwise_sum(intersection) + smooth  # (C,)
    denominator = probs + onehot_target  # (N, C, ...)
    denominator = _channelwise_sum(denominator) + smooth + eps  # (C,)
    loss_per_channel = 1 - (numerator / denominator)  # (C,)
    weighted_loss_per_channel = weight * loss_per_channel  # (C,)
    return weighted_loss_per_channel.mean()  # ()


class DiceLoss(torch.nn.Module):
    """Generalized Dice Loss, as described in https://arxiv.org/abs/1707.03237.

    Works for n-dimensional data. Assuming that the ``output`` tensor to be
    compared to the ``target`` has the shape (N, C, D, H, W), the ``target``
    can either have the same shape (N, C, D, H, W) (one-hot encoded) or
    (N, D, H, W) (with dense class indices, as in
    ``torch.nn.CrossEntropyLoss``). If the latter shape is detected, the
    ``target`` is automatically internally converted to a one-hot tensor
    for loss calculation.

    Args:
        apply_softmax: If ``True``, a softmax operation is applied to the
            ``output`` tensor before loss calculation. This is necessary if
            your model does not already apply softmax as the last layer.
            If ``False``, ``output`` is assumed to already contain softmax
            probabilities.
        weight: Weight tensor for class-wise loss rescaling.
            Has to be of shape (C,). If ``None``, classes are weighted equally.
        smooth: Smoothing term that is added to both the numerator and the
            denominator of the dice loss formula.
    """
    def __init__(
            self,
            apply_softmax: bool = True,
            weight: Optional[torch.Tensor] = None,
            smooth: float = 0.
    ):
        super().__init__()
        if apply_softmax:
            self.softmax = torch.nn.Softmax(dim=1)
        else:
            self.softmax = lambda x: x  # Identity (no softmax)
        self.dice = dice_loss
        if weight is None:
            weight = torch.tensor(1.)
        self.register_buffer('weight', weight)
        self.smooth = smooth

    def forward(self, output, target):
        probs = self.softmax(output)
        return self.dice(probs, target, weight=self.weight, smooth=self.smooth)


# TODO: There is some low-hanging fruit for performance optimization
class FixMatchSegLoss(nn.Module):
    """Self-supervised loss for semi-supervised semantic segmentation training,
    very similar to the :math:`l_u` loss proposed in
    FixMatch (https://arxiv.org/abs/2001.07685).

    The main difference to FixMatch is the kind of augmentations that are used
    for consistency regularization. In FixMatch, so-called
    "strong augmentations" are applied to the (already "weakly augmented")
    inputs. Most of these strong augmentations only work for image-level
    classification.
    In ``FMSegLoss``, only simple, easily reversible geometric augmentations
    are used currently
    (random xy(z) flipping and random xy rotation in 90 degree steps).
    TODO: Add more augmentations

    This loss combines two different well-established semi-supervised learning
    techniques:

    - consistency regularization: consistency (equivariance) against random
      flipping and random rotation augmentatations is enforced
    - pseudo-label training: model argmax predictions are treated as targets
      for a pseudo-supervised cross-entropy training loss
      This only works for settings where argmax makes sense (not suitable for
      regression) and can be disabled with ``enable_psuedo_label=False``.

    Args:
        model: Neural network model to be trained.
        scale: Scalar factor to be multiplied with the loss to adjust its
            magnitude. (If this loss is combined with a standard supervised
            cross entropy, ``scale`` corresponds to the lambda_u hyperparameter
            in FixMatch
        enable_pseudo_label: If ``enable_pseudo_label=True``, the inner loss is
            the cross entropy between the argmax pseudo label tensor computed
            from the weakly augmented input and the softmax model output on the
            strongly augmented input. Since this internally uses
            ``torch.nn.CrossEntropyLoss``, the ``model`` is expected to give
            raw, unsoftmaxed outputs.
            This only works for settings where computing the argmax and softmax
            on the outputs makes sense (so classification, not regression).
            If ``enable_pseudo_label=False``, a mean squared error regression
            loss is computed directly on the difference between the two model
            outputs, without computing or using pseudo-labels.
            In this case, the loss is equivalent to the ``R`` loss proposed in
            "Transformation Consistent Self-ensembling Model for
            Semi-supervised Medical Image Segmentation"
            (https://arxiv.org/abs/1903.00348).
            This non-pseudo-label variant of the loss can also be used for
            pixel-level regression training.
        confidence_thresh: (Only applies if ``enable_pseudo_label=True``.)
            The confidence threshold that determines how
            confident the model has to be in each output element's
            classification for it to contribute to the loss. All output
            elements where none of the softmax class probs exceed this
            threshold are masked out from the loss calculation and the resulting
            loss is set to 0. In the FixMatch paper, this hyperparameter is
            called tau.
        ce_weight: (Only applies if ``enable_pseudo_label=True``.)
            Class weight tensor for the inner cross-entropy loss. Should be
            the same as the weight for the supervised cross-entropy loss.
    """

    def __init__(
            self,
            model: nn.Module,
            scale: float = 1.,
            enable_pseudo_label: bool = True,
            confidence_thresh: float = 0.9,
            ce_weight=None
    ):
        super().__init__()
        self.model = model
        self.scale = scale
        self.enable_pseudo_label = enable_pseudo_label
        self.confidence_thresh = confidence_thresh

        if self.enable_pseudo_label:
            self.criterion = nn.CrossEntropyLoss(weight=ce_weight, ignore_index=-100)
        else:
            self.criterion = nn.MSELoss()

    @staticmethod
    def get_random_augmenters(
            ndim: int
    ) -> Tuple[Callable[[torch.Tensor], torch.Tensor], Callable[[torch.Tensor], torch.Tensor]]:
        """Produce a pair of functions ``augment, reverse_augment``, where
        the ``augment`` function applies a random augmentation to a torch
        tensor and the ``reverse_augment`` function performs the reverse
        aumentations if applicable (i.e. for geometrical transformations)
        so pixel-level loss calculation is still correct).

        Note that all augmentations are performed on the compute device that
        holds the input, so generally on the GPU.
        """

        # Random rotation angle (in 90 degree steps)
        k90 = torch.randint(0, 4, ()).item()
        # Get a random selection of spatial dims (ranging from [] to [2, 3, ..., example.ndim - 1]
        flip_dims_binary = torch.randint(0, 2, (ndim - 2,))
        flip_dims = (torch.nonzero(flip_dims_binary, as_tuple=False).squeeze(1) + 2).tolist()

        @torch.no_grad()
        def augment(x: torch.Tensor) -> torch.Tensor:
            x = torch.rot90(x, +k90, (-1, -2))
            if len(flip_dims) > 0:
                x = torch.flip(x, flip_dims)

            # # Uncomment to enable additional random brightness and contrast augmentations
            # contrast_std = 0.1
            # brightness_std = 0.1
            # a = torch.randn(x.shape[:2], device=x.device, dtype=x.dtype) * contrast_std + 1.0
            # b = torch.randn(x.shape[:2], device=x.device, dtype=x.dtype) * brightness_std
            # for n in range(x.shape[0]):
            #     for c in range(x.shape[1]):
            #         # Formula based on tf.image.{adjust_contrast,adjust_brightness}
            #         # See https://www.tensorflow.org/api_docs/python/tf/image
            #         m = torch.mean(x[n, c])
            #         x[n, c] = a[n, c] * (x[n, c] - m) + m + b[n, c]

            # # Uncomment to enable additional additive gaussian noise augmentations
            # agn_std = 0.1
            # x.add_(torch.randn_like(x).mul_(agn_std))

            return x

        @torch.no_grad()
        def reverse_augment(x: torch.Tensor) -> torch.Tensor:
            if len(flip_dims) > 0:  # Check is necessary only on cuda
                x = torch.flip(x, flip_dims)
            x = torch.rot90(x, -k90, (-1, -2))
            return x

        return augment, reverse_augment

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        augment, reverse_augment = self.get_random_augmenters(ndim=inp.ndim)
        aug = augment(inp)
        out = self.model(inp)
        aug_out = self.model(aug)
        aug_out_reversed = reverse_augment(aug_out)

        if self.enable_pseudo_label:
            with torch.no_grad():
                # We need softmax outputs for thresholding
                out = torch.softmax(out, 1)
                omax, pseudo_label = torch.max(out, dim=1)
                # Ignore loss on outputs that are lower than the confidence threshold
                mask = omax < self.confidence_thresh
                # Assign special ignore value to all masked elements
                pseudo_label[mask] = self.criterion.ignore_index
            loss = self.criterion(aug_out_reversed, pseudo_label)
        else:
            loss = self.criterion(aug_out_reversed, out)
        scaled_loss = self.scale * loss
        return scaled_loss


# TODO: Rename and clean up
def norpf_dice_loss(probs, target, weight=1., class_weight=1.):
    # Probs need to be softmax probabilities, not raw network outputs
    tsh, psh = target.shape, probs.shape

    if tsh == psh:  # Already one-hot
        onehot_target = target.to(probs.dtype)
    elif tsh[0] == psh[0] and tsh[1:] == psh[2:]:  # Assume dense target storage, convert to one-hot
        onehot_target = torch.zeros_like(probs)
        onehot_target.scatter_(1, target.unsqueeze(1), 1)
    else:
        raise ValueError(
            f'Target shape {target.shape} is not compatible with output shape {probs.shape}.'
        )
    # if weight is None:
    #     weight = torch.ones(probs.shape[0], dtype=probs.dtype)  # (C,)
    # if ignore_index is not None:
    #     weight[:, ignore_index] = 0.

    if weight.sum() == 0:
        return probs.sum() * 0

    ignore_mask = (1 - onehot_target[0][-1]).view(1,1,*probs.shape[2:]) # inverse ignore

    bg_probs = 1 - probs
    bg_target = 1 - onehot_target
    global_weight = (class_weight > 0).type(probs.dtype)
    positive_target_mask = (weight.view(1,-1,1,1,1) * onehot_target)[0][1:-1].sum(dim=0).view(1,1,*probs.shape[2:]) # weighted targets w\ background and ignore
    weight = weight * global_weight
    dense_weight = weight.view(1,-1,1,1,1)
    target_mask_empty = ((positive_target_mask * ignore_mask).sum(dim=(0,2,3,4)) == 0).type(probs.dtype)
    target_empty = ((onehot_target * ignore_mask).sum(dim=(0,2,3,4)) == 0).type(probs.dtype)
    bg_target_empty = ((bg_target * ignore_mask).sum(dim=(0,2,3,4)) == 0).type(probs.dtype)
    # complete background for weighted classes and target of weighted classes as background for unweighted classes
    needs_positive_target_mark = (dense_weight.sum() == 0).type(probs.dtype)
    bg_mask = torch.ones_like(bg_probs) * dense_weight + needs_positive_target_mark * positive_target_mask * global_weight.view(1,-1,1,1,1)

    # make num/denom 1 for unweighted classes and classes with no target
    intersection = probs * onehot_target * ignore_mask * dense_weight  # (N, C, ...)
    intersection2 = bg_probs * bg_target * ignore_mask * bg_mask  # (N, C, ...)
    denominator = (probs + onehot_target) * ignore_mask * dense_weight  # (N, C, ...)
    denominator2 = (bg_probs + bg_target) * ignore_mask * bg_mask  # (N, C, ...)
    numerator = 2 * class_weight * _channelwise_sum(intersection)  # (C,)
    numerator2 = 2 * _channelwise_sum(intersection2)  # (C,)
    denominator = class_weight * _channelwise_sum(denominator)  # (C,)
    denominator2 = _channelwise_sum(denominator2)  # (C,)

    no_tp = (numerator == 0).type(probs.dtype)
    # workarounds for divide by zero
    # unweighted classes get DSC=1
    numerator += (1 - weight)
    denominator += (1 - weight)
    bg_mask_empty = ((bg_mask).sum(dim=(0,2,3,4)) == 0).type(probs.dtype)
    numerator2 *= 1 - bg_mask_empty
    numerator2 += bg_mask_empty
    denominator2 *= 1 - bg_mask_empty
    denominator2 += bg_mask_empty
    # when there is no target, DSC has no meaning and is therefore set to 1 as well
    numerator *= 1 - target_empty
    numerator += target_empty
    denominator *= 1 - target_empty
    denominator += target_empty
    numerator2 *= 1 - bg_target_empty
    numerator2 += bg_target_empty
    denominator2 *= 1 - bg_target_empty
    denominator2 += bg_target_empty

    if (denominator == 0).sum() > 0 or (denominator2 == 0).sum() > 0:
        print(denominator, denominator2)
        import IPython
        IPython.embed()

    #numerator2 += (numerator2 == 0).type(numerator2.dtype) # no tp background
    #denominator2 += (denominator2 == 0).type(denominator2.dtype) # 100% tp foreground
    #denominator += (denominator == 0).type(denominator.dtype) # ???¿¿¿

    #loss_per_channel = 1 - numerator / denominator  # (C,)
    #loss_per_channel = 1 - ((numerator * denominator2 + denominator * numerator2) / (2 * denominator * denominator2))  # (C,)

    #loss_per_channel = 1 - (numerator * denominator2 + denominator * numerator2) / (2 * denominator * denominator2) # (C,)
    loss_per_channel = 1 + no_tp - (numerator/denominator + no_tp * numerator2/denominator2) # (C,)

    #loss_per_channel = 1 - (numerator + numerator2 + target_mask_empty * (1 - weight)) / (denominator + denominator2 + target_mask_empty * (1 - weight)) # (C,)
    #weighted_loss = 1 - (numerator.sum() + numerator2.sum())/(denominator.sum() + denominator2.sum())
    #weighted_loss = 1 - (numerator[1:-1].sum() + numerator2[1:-1].sum()) / (denominator[1:-1].sum() + denominator2[1:-1].sum()) # (C,)

    #weighted_loss = (weight[:-1] * loss_per_channel[:-1]).sum() / weight[:-1].sum()  # (C,)

    # normalize loss to [0, 1]
    #weighted_loss = loss_per_channel[1:-1].sum() / ((loss_per_channel[1:-1] > 0).sum() + (loss_per_channel[1:-1].sum() == 0))  # (C,)
    #weighted_loss *= 2
    #print(loss_per_channel)
    weighted_loss = loss_per_channel[1:-1].sum() / (class_weight[1:-1] > 0).sum()
    #weighted_loss = loss_per_channel[1:-1].sum() 

    if torch.isnan(weighted_loss).sum() or (weighted_loss > 1).sum():
    #if torch.isnan(weighted_loss):
        print(loss_per_channel)
        import IPython
        IPython.embed()

    return weighted_loss  # ()


# TODO: Rename and clean up
class NorpfDiceLoss(torch.nn.Module):
    """Generalized Dice Loss, as described in https://arxiv.org/abs/1707.03237,

    Works for n-dimensional data. Assuming that the ``output`` tensor to be
    compared to the ``target`` has the shape (N, C, D, H, W), the ``target``
    can either have the same shape (N, C, D, H, W) (one-hot encoded) or
    (N, D, H, W) (with dense class indices, as in
    ``torch.nn.CrossEntropyLoss``). If the latter shape is detected, the
    ``target`` is automatically internally converted to a one-hot tensor
    for loss calculation.

    Args:
        apply_softmax: If ``True``, a softmax operation is applied to the
            ``output`` tensor before loss calculation. This is necessary if
            your model does not already apply softmax as the last layer.
            If ``False``, ``output`` is assumed to already contain softmax
            probabilities.
        weight: Weight tensor for class-wise loss rescaling.
            Has to be of shape (C,). If ``None``, classes are weighted equally.
    """
    def __init__(self, apply_softmax=True, weight=torch.tensor(1.), class_weight=torch.tensor(1.)):
        super().__init__()
        if apply_softmax:
            self.softmax = torch.nn.Softmax(dim=1)
        else:
            self.softmax = lambda x: x  # Identity (no softmax)
        self.dice = norpf_dice_loss
        self.register_buffer('weight', weight)
        self.register_buffer('class_weight', class_weight)

    def forward(self, output, target):
        probs = self.softmax(output)
        return self.dice(probs, target, weight=self.weight, class_weight=self.class_weight)


class LovaszLoss(torch.nn.Module):
    """https://arxiv.org/abs/1705.08790"""
    def __init__(self, apply_softmax=True):
        super().__init__()
        if apply_softmax:
            self.softmax = torch.nn.Softmax(dim=1)
        else:
            self.softmax = lambda x: x  # Identity (no softmax)
        # lovasz_softmax works on softmax probs, so we still have to apply
        #  softmax before passing probs to it
        self.lovasz = lovasz_softmax

    def forward(self, output, target):
        probs = self.softmax(output)
        return self.lovasz(probs, target)


class ACLoss(torch.nn.Module):
    """Active Contour loss
    http://openaccess.thecvf.com/content_CVPR_2019/papers/Chen_Learning_Active_Contour_Models_for_Medical_Image_Segmentation_CVPR_2019_paper.pdf

    Supports 2D and 3D data, as long as all spatial dimensions have the same
    size and there are only two output channels.

    Modifications:
    - Using mean instead of sum for reductions to avoid size dependency.
    - Instead of the proposed λ loss component weighting (which leads to
      exploding loss magnitudes for high λ values), a relative weight
      ``region_weight`` is used to balance the components:
      ``ACLoss = (1 - region_weight) * length_term + region_weight * region_term``
    """

    def __init__(self, num_classes: int, region_weight: float = 0.5):
        assert 0. <= region_weight <= 1., 'region_weight must be between 0 and 1'
        self.num_classes = num_classes
        self.region_weight = region_weight
        super().__init__()

    @staticmethod
    def get_length(output):
        if output.ndim == 4:
            dy = output[:, :, 1:, :] - output[:, :, :-1, :]  # y gradient (B, C, H-1, W)
            dx = output[:, :, :, 1:] - output[:, :, :, :-1]  # x gradient (B, C, H, W-1)
            dy = dy[:, :, 1:, :-2] ** 2  # (B, C, H-2, W-2)
            dx = dx[:, :, :-2, 1:] ** 2  # (B, C, H-2, W-2)
            delta_pred = torch.abs(dy + dx)
        elif output.ndim == 5:
            assert output.shape[3] == output.shape[4], 'All spatial dims must have the same size'
            dz = output[:, :, 1:, :, :] - output[:, :, :-1, :, :]  # z gradient (B, C, D-1, H, W)
            dy = output[:, :, :, 1:, :] - output[:, :, :, :-1, :]  # h gradient (B, C, D, H-1, W)
            dx = output[:, :, :, :, 1:] - output[:, :, :, :, :-1]  # w gradient (B, C, D, H, W-1)
            dz = dz[:, :, 1:, :-2, :-2] ** 2  # (B, C, D-2, H-2, W-2)
            dy = dy[:, :, :-2, 1:, :-2] ** 2  # (B, C, D-2, H-2, W-2)
            dx = dx[:, :, :-2, :-2, 1:] ** 2  # (B, C, D-2, H-2, W-2)
            delta_pred = torch.abs(dz + dy + dx)
        length = torch.mean(torch.sqrt(delta_pred + 1e-6))
        return length

    @staticmethod
    def get_region(output, target):
        region_in = torch.mean(output * (target - 1.) ** 2.)
        region_out = torch.mean((1 - output) * target ** 2.)
        return region_in + region_out

    def forward(self, output, target):
        assert output.shape[2] == output.shape[3], 'All spatial dims must have the same size'
        if target.ndim == output.ndim - 1:
            target = torch.nn.functional.one_hot(target, self.num_classes).transpose(1, -1)

        length_term = self.get_length(output) if self.region_weight < 1. else 0.
        region_term = self.get_region(output, target) if self.region_weight > 0. else 0.
        loss = (1 - self.region_weight) * length_term + self.region_weight * region_term
        return loss


class MixedCombinedLoss(torch.nn.Module):
    """
    Defines a loss function as a weighted sum of combinable loss criteria for multi-class classification with only
    single class ground truths.

    For each voxel, we construct a 2 channel output after the softmax:
     channel 0: background (actual background + all but one classes) = (1-channel 1)
     channel 1: foreground (the one class to which the target corresponds)

    Args:
        class_weight: a manual rescaling weight given to each
            class.
        criteria: List of loss criterion modules that should be combined.
        criteria_weight: Weight assigned to the individual loss criteria (in the same
            order as ``criteria``).
        device: The device on which the loss should be computed. This needs
            to be set to the device that the loss arguments are allocated on.
        eps:
    """

    def __init__(self, class_weight, criteria, criteria_weight, device, eps=1e-10, **kwargs):
        super(MixedCombinedLoss, self).__init__()

        self.softmax = torch.nn.Softmax(dim=1)
        self.class_weight = class_weight

        self.criteria = torch.nn.ModuleList(criteria)
        self.device = device

        self.eps = eps

        if criteria_weight is None:
            weight = torch.ones(len(criteria))
        else:
            weight = torch.as_tensor(criteria_weight, dtype=torch.float32)
            assert weight.shape == (len(criteria),)
        self.register_buffer('weight', weight.to(self.device))

    def forward(self, output_direct, target, target_class):

        assert all([len(torch.unique(target_sample[0])) <= 2 for target_sample in
                    target])  # background and that class for each sample in the batch
        modified_target = torch.zeros_like(target)
        modified_target[target != 0] = 1

        logit_max = output_direct.max(axis=1)[0].unsqueeze(1)
        output_shifted = output_direct - logit_max  # subtract max for numerical stability

        # for dice
        softmax_output = self.softmax(output_shifted)
        softmax_output = (1 - self.eps) * softmax_output + self.eps  # eps for numerical stability
        softmax_output = softmax_output[(range(softmax_output.shape[0]), target_class)].unsqueeze(1)
        softmax_output = torch.cat([1 - softmax_output, softmax_output], dim=1)

        # for crossentropy: compute in log softmax for the two classes for numerical stability
        exp_output = output_shifted.exp()
        exp_output_sum = exp_output.sum(axis=1).unsqueeze(1)
        exp_output_sum_log = exp_output_sum.log()
        # foreground i: log_softmax_i = -log(softmax(x)[i]) = x[i] - log(\sum_j exp(x[j]))
        log_softmax_i = output_shifted - exp_output_sum_log

        num_classes = output_direct.shape[1]
        idx = [np.arange(num_classes) != i for i in range(num_classes)]  # todo: only compute for target class
        exp_output_sum_minus_i = torch.stack([exp_output[:, idx[k]].sum(axis=1) for k in range(num_classes)], dim=1)
        # assert torch.allclose(exp_output_sum_minus_i, exp_output_sum - exp_output)
        # background -i: log_softmax_minus_i = log(1-softmax(x)[i]) = log(\sum_j!=i exp(x[j])) - log(\sum_j exp(x[j]))
        log_softmax_minus_i = (exp_output_sum_minus_i).log() - exp_output_sum_log

        log_softmax_i_output = log_softmax_i[range(softmax_output.shape[0]), target_class].unsqueeze(1)
        log_softmax_minus_i_output = log_softmax_minus_i[range(softmax_output.shape[0]), target_class].unsqueeze(1)
        log_softmax_output = torch.cat([log_softmax_minus_i_output, log_softmax_i_output], dim=1)

        loss = torch.tensor(0., device=softmax_output.device)
        for crit, crit_weight in zip(self.criteria, self.weight):
            for i in range(softmax_output.shape[0]):  # todo: not element-wise; process full batch
                if isinstance(crit, torch.nn.NLLLoss):
                    crit_loss = crit(log_softmax_output[i].unsqueeze(0), modified_target[i].unsqueeze(0)).mean()
                elif isinstance(crit, DiceLoss):
                    crit_loss = crit(softmax_output[i].unsqueeze(0), modified_target[i].unsqueeze(0))
                else:
                    raise NotImplementedError()
                loss += crit_loss * crit_weight * self.class_weight[target_class[i]]
        return loss

##### ALTERNATIVE VERSIONS OF DICE LOSS #####

# Version with features that are untested and currently not needed
# Based on https://discuss.pytorch.org/t/one-hot-encoding-with-autograd-dice-loss/9781/5
def __dice_loss_with_cool_extra_features(output, target, weights=None, ignore_index=None):
    eps = 0.0001

    encoded_target = torch.zeros_like(output)
    if ignore_index is not None:
        mask = target == ignore_index
        target = target.clone()
        target[mask] = 0
        encoded_target.scatter_(1, target.unsqueeze(1), 1)
        mask = mask.unsqueeze(1).expand_as(encoded_target)
        encoded_target[mask] = 0
    else:
        encoded_target.scatter_(1, target.unsqueeze(1), 1)

    if weights is None:
        weights = 1

    intersection = output * encoded_target
    numerator = 2 * _channelwise_sum(intersection)
    denominator = output + encoded_target

    if ignore_index is not None:
        denominator[mask] = 0
    denominator = _channelwise_sum(denominator) + eps
    loss_per_channel = weights * (1 - (numerator / denominator))

    return loss_per_channel.sum() / output.shape[1]


# Very simple version. Only for binary classification. Just included for testing.
# Note that the smooth value is set to 0 and eps is introduced instead, to make it comparable.
# Based on https://github.com/pytorch/pytorch/issues/1249#issuecomment-305088398
def __dice_loss_binary(output, target, smooth=0, eps=0.0001):
    onehot_target = torch.zeros_like(output)
    onehot_target.scatter_(1, target.unsqueeze(1), 1)

    iflat = output.view(-1)
    tflat = onehot_target.view(-1)
    intersection = (iflat * tflat).sum()

    return 1 - ((2 * intersection + smooth) /
                (iflat.sum() + tflat.sum() + smooth + eps))



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Philipp Schubert, Martin Drawitsch
import copy
import itertools
import logging
import os
import time
import zipfile
from collections import OrderedDict
from typing import Optional, Tuple, Union, Callable, Sequence
from pathlib import Path

import numpy as np
import torch
from torch import nn
from tqdm import tqdm

from elektronn3.data import utils


# TODO: It's confusing that tiled_apply expects out_shape to include the N dim, but
#     Predictor has a parameter with the same name but which doesn't include N.

logger = logging.getLogger('elektronn3log')

# Alias for type hinting
Transform = Callable[
    [np.ndarray, Optional[np.ndarray]],
    Tuple[np.ndarray, Optional[np.ndarray]]
]


def _extend_nc(spatial_slice: Sequence[slice]) -> Tuple[slice, ...]:
    """Extend a spatial slice ([D,] H, W) to also include the non-spatial (N, C) dims."""
    # Slice everything in N and C dims (equivalent to [:, :] in direct notation)
    nonspatial_slice = [slice(None)] * 2
    return tuple(nonspatial_slice + list(spatial_slice))


# TODO Fix and document out_shape change for argmax outputs

def tiled_apply(
        func: Callable[[torch.Tensor], torch.Tensor],
        inp: torch.Tensor,
        tile_shape: Sequence[int],
        overlap_shape: Sequence[int],
        offset: Optional[Sequence[int]],
        out_shape: Sequence[int],
        verbose: bool = False
) -> torch.Tensor:
    """Splits a tensor into overlapping tiles and applies a function on them independently.

    Each tile of the output results from applying a callable ``func`` on an
    input tile which is sliced from a region that has the same center but a
    larger extent (overlapping with other input regions in the vicinity).
    Input tensors are also padded with zeros at the boundaries according to
    the ``overlap_shape`` to enable consistent tile shapes.

    The overlapping behavior prevents imprecisions of CNNs (and image
    processing algorithms in general) that appear near the boundaries of
    inner tiles when applying them on a tiled representation of the input.

    By default this function assumes that ``inp.shape[2:] == func(inp).shape[2:]``,
    i.e. that the function keeps the spatial shape unchanged.
    If ``func`` reduces the spatial shape (e.g. by performing valid convolutions)
    and its output is centered w.r.t. the input, you should specify this shape
    offset in the ``offset`` parameter. This is the same offset that
    :py:class:`elektronn3.data.cnndata.PatchCreator` expects.

    It can run on GPU or CPU transparently, depending on the device that
    ``inp`` is allocated on.

    Although this function is mainly intended for the purpose of neural network
    inference, ``func`` doesn't have to be a neural network but can be
    any ``Callable[[torch.Tensor], torch.Tensor]`` that operates on n-dimensional
    image data of shape (N, C, ...) and preserves spatial shape or has a
    constant ``offset``.
    ("..." is a placeholder for the spatial dimensions, so for example
    H(eight) and W(idth).)


    Args:
        func: Function to be applied on input tiles. Usually this is a neural
            network model.
        inp: Input tensor, usually of shape (N, C, [D,], H, W).
            n-dimensional tensors of shape (N, C, ...) are supported.
        tile_shape: Spatial shape of the output tiles to use for inference.
        overlap_shape: Spatial shape of the overlap by which input tiles are
            extended w.r.t. the output ``tile_shape``.
        offset: Determines the offset by which the output contents are shifted
            w.r.t. the inputs by ``func``.
            This should generally be set to half the spatial shape difference
            between inputs and outputs:

            >>> in_sh = np.array(inp.shape[2:])
            >>> out_sh = np.array(func(inp).shape[2:])
            >>> offset = (in_sh - out_sh) // 2

        out_shape: Expected shape of the output tensor that would result from
            applying ``func`` to ``inp`` (``func(inp).shape``).
            It doesn't just refer to spatial shape, but to the actual tensor
            shape including N and C dimensions.
            Note: ``func(inp)`` is never actually executed – ``out_shape`` is
            merely used to pre-allocate the output tensor so it can be filled
            later.
        verbose: If ``True``, a progress bar will be shown while iterating over
            the tiles.

    Returns:
        Output tensor, as a torch tensor of the same shape as the input tensor.
    """
    if not (inp.dim() - 2 == len(tile_shape) == len(overlap_shape)):
        raise ValueError(
            f'ndims of tile shape ({len(tile_shape)}) and overlap shape '
            f'({len(overlap_shape)}) don\'t match input shape ndim - 2'
            f'({inp.dim() - 2}).'
        )
    if not np.all(np.mod(out_shape[2:], tile_shape) == 0):
        raise ValueError(
            f'spatial out shape[2:] {tuple(out_shape[2:])} has to be divisible '
            f'by tile_shape {tile_shape}.'
        )
    if offset is not None:
        offset = np.array(offset)
    inp_shape = np.array(inp.shape)
    out = None
    out_shape = np.array(out_shape)
    tile_shape = np.array(tile_shape)
    overlap_shape = np.array(overlap_shape)

    if not np.array_equal(out_shape[2:], inp_shape[2:]): # input is already padded
        inp_padded = inp
    else:
        # Create padded input with overlap
        padded_shape = inp_shape + np.array((0, 0, *overlap_shape * 2))
        logger.info(f'additional input padding to {padded_shape}')
        inp_padded = torch.zeros(tuple(padded_shape), dtype=inp.dtype)

        padslice = _extend_nc(
            [slice(l, h) for l, h in zip(overlap_shape, padded_shape[2:] - overlap_shape)]
        )
        inp_padded[padslice] = inp

        crop_low_corner = overlap_shape.copy()
        crop_high_corner = tile_shape + overlap_shape
        # Used to crop the output tile to the relevant, unpadded region
        #  that will be written to the final output
        final_crop_slice = _extend_nc([slice(l, h) for l, h in zip(crop_low_corner, crop_high_corner)])
    if offset is not None: # no cropping necessary for valid conv
        final_crop_slice = None
    del inp

    tiles = np.ceil(out_shape[2:] / tile_shape).astype(int)
    num_tiles = np.prod(tiles)

    tile_ranges = [range(t) for t in tiles]
    # TODO: Handle fractional inputshape-to-tile ratio
    pbar = tqdm(
        itertools.product(*tile_ranges), 'Predicting',
        total=num_tiles, disable=not verbose, dynamic_ncols=True
    )
    for tile_pos in pbar:
        tile_pos = np.array(tile_pos)
        # Calculate corner coordinates of the current output tile
        out_low_corner = tile_shape * tile_pos
        out_high_corner = tile_shape * (tile_pos + 1)

        # Note: To understand why the input corners are chosen in this
        #  particular way, it helps to draw this on paper in the 1d case,
        #  representing the input tensor as a line and slicing it into
        #  input and output tiles, where input tiles have a certain overlap
        #  (note that input and output coordinates exist in two different
        #  coordinate systems: input corner coordinates are shifted "right" by
        #  the ``overlap_shape`` w.r.t. the output corner coordinate system
        #  due to the initial padding.
        inp_low_corner = out_low_corner.copy()
        inp_high_corner = out_high_corner.copy() + 2 * overlap_shape

        assert np.all(np.less_equal(inp_high_corner, inp_padded.shape[2:])), inp_high_corner
        # Slice only the current tile region in ([D,] H, W) dims
        # Slice input with overlap
        inp_slice = _extend_nc([slice(l, h) for l, h in zip(inp_low_corner, inp_high_corner)])
        # Output slice without overlap (this is the region where the current
        #  inference result will be stored)
        out_slice = _extend_nc([slice(l, h) for l, h in zip(out_low_corner, out_high_corner)])
        inp_tile = inp_padded[inp_slice].contiguous()
        out_tile = func(inp_tile, final_crop_slice)

        # Slice the relevant tile_shape-sized region out of the model output
        #  so it can be written to the final output
        # Since out is a CPU tensor, out[out_slice] assignments below implicitly copy data to CPU
        if out is None:
            out = torch.empty(out_shape.tolist(), dtype=out_tile.dtype)
        out[out_slice] = out_tile

    return out


class Argmax(nn.Module):
    def __init__(self, dim=1, unsqueeze=True):
        super().__init__()
        self.dim = dim
        self.unsqueeze = unsqueeze

    def forward(self, x):
        argmax = torch.argmax(x, self.dim)
        if self.unsqueeze:  # Restore C dim as a workaround for unified slicing pattern in tiled_apply()
            argmax.unsqueeze_(1)
        return argmax


class FlipAugment:
    def __init__(self, dims):
        self.dims = tuple(np.array(dims) + 2)  # Dim offset to skip (N, C) dims

    def forward(self, inp):
        return torch.flip(inp, dims=self.dims)

    def backward(self, inp):
        return self.forward(inp)


# TODO
# class Rot90Augment:
#     def __init__(self, k, dims):
#         self.k = k
#         self.dims = dims
#
#     def forward(self, inp):
#         return torch.rot90(inp, k=self.k, dims=self.dims)
#
#     def backward(self, inp):
#         return torch.rot90(inp, k=-self.k, dims=self.dims)


DEFAULT_AUGMENTATIONS_3D = [  # Flip every dim
    FlipAugment(dims)
    for dims in [(0,), (1,), (0, 1), (2,), (0, 2), (1, 2), (0, 1, 2)]
]
DEFAULT_AUGMENTATIONS_2D = DEFAULT_AUGMENTATIONS_3D[:3]  # Limit flips to first 2 dims


class Predictor:
    """Class to perform inference using a ``torch.nn.Module`` object either
    passed directly or loaded from a file.

    If both ``tile_shape`` and ``overlap_shape`` are ``None``, input tensors
    are fed directly into the ``model`` (best for scalar predictions,
    medium-sized 2D images or very small 3D images).
    If you define ``tile_shape`` and ``overlap_shape``, these are used to
    slice a large input into smaller overlapping tiles and perform predictions
    on these tiles independently and later put the output tiles together into
    one dense tensor without overlap again. Use this features if your model
    has spatially interpretable (dense) outputs and if passing one input sample
    to the ``model`` would result in an out-of-memory error. For more details
    on this tiling mode, see
    :py:meth:`elektronn3.inference.inference.tiled_apply()`.

    Args:
        model: Network model to be used for inference.
            The model can be passed as an ``torch.nn.Module``, or as a path
            to either a model file or to an elektronn3 save directory:

            - If ``model`` is a ``torch.nn.Module`` object, it is used
              directly.
            - If ``model`` is a path (string) to a serialized TorchScript
              module (.pts), it is loaded from the file and mapped to the
              specified ``device``.
            - If ``model`` is a path (string) to a pickled PyTorch module (.pt)
              (**not** a pickled ``state_dict``), it is loaded from the file
              and mapped to the specified ``device`` as well.
        state_dict_src: Path to ``state_dict`` file (.pth) or loaded
            ``state_dict`` or ``None``. If not ``None``, the ``state_dict`` of
            the ``model`` is replaced with it.
        device: Device to run the inference on. Can be a ``torch.device`` or
            a string like ``'cpu'``, ``'cuda:0'`` etc.
            If not specified (``None``), available GPUs are automatically used;
            the CPU is used as a fallback if no GPUs can be found.
        batch_size: Maximum batch size with which to perform
            inference. In general, a higher ``batch_size`` will give you
            higher prediction speed, but prediction will consume more
            GPU memory. Reduce the ``batch_size`` if you run out of memory.
            If this is ``None`` (default), the input batch size is used
            as the prediction batch size.
        tile_shape: Spatial shape of the output tiles to use for inference.
            The spatial shape of the input tensors has to be divisible by
            the ``tile_shape``.
        overlap_shape: Spatial shape of the overlap by which input tiles are
            extended w.r.t. the ``tile_shape`` of the resulting output tiles.
            The ``overlap_shape`` should be close to the effective receptive
            field of the network architecture that's used for inference.
            Note that ``tile_shape + 2 * overlap`` needs to be a valid
            input shape for the inference network architecture, so
            depending on your network architecture (especially pooling layers
            and strides), you might need to adjust your ``overlap_shape``.
            If your inference fails due to shape issues, as a rule of thumb,
            try adjusting your ``overlap_shape`` so that
            ``tile_shape + 2 * overlap`` is divisible by 16 or 32.

            If ``offset`` (see below) is not ``None``, ``overlap_shape``
            can't be specified but it is configured automatically.
        offset:  Shape of the offset by which each the output tiles are smaller
            than the input tiles
            on each side. This applies for networks using valid convolutions.
            If ``offset`` is specified, ``overlap_shape`` (see above) can't
            be specified but is configured automatically.
        out_shape: Expected shape of the output tensor.
            It doesn't just refer to spatial shape, but to the actual tensor
            shape of one sample, including the channel dimension C, but
            **excluding** the batch dimension N.
            Note: ``model(inp)`` is never actually executed if tiling is used
            – ``out_shape`` is merely used to pre-allocate the output tensor so
            it can be filled later.
            If you know how many channels your model output has
            (``out_channels``) and if your model
            preserves spatial shape, you can easily calculate ``out_shape``
            yourself as follows:

            >>> out_channels: int = ?  # E.g. for binary classification it's 2
            >>> out_shape = (out_channels, *inp.shape[2:])
        out_dtype: torch dtype that the output will be cast to
        float16: If ``True``, deploy the model in float16 (half) precision.
        apply_softmax: If ``True``
            (default), a softmax operator is automatically appended to the
            model, in order to get probability tensors as inference outputs
            from networks that don't already apply softmax.
        apply_argmax: If ``True``, the argmax of the model output is computed
            and returned instead of the class score tensor.  This can be used
            for classification if you are only interested in the final argmax
            classification. This option can speed up predictions.
            Note that since argmax is not influenced by softmax,
            ``apply_softmax`` can be safely disabled if ``apply_argmax`` is
            ``True``, even if the model was trained with a softmax loss.
        transform: Transformation function to be applied to inputs before
            performing inference. The primary use of this is for normalization.
            Make sure to use the same normalization parameters for inference as
            the ones that were used for training of the ``model``.
            See :py:mod:`elektronn3.data.transforms`. for some implementations.
            For pure input normalization you can use this template::

            >>> from elektronn3.data import transforms
            >>> # m, s are mean, std of the inputs the model was trained on
            >>> transform = transforms.Normalize(mean=m, std=s)
        augmentations: List of test-time augmentations or integer that
            specifies the number of different flips to be performed as test-
            time augmentations.
        strict_shapes: If ``False`` (default), force the ``output_shape`` to be
            a multiple of the ``tile_shape`` by padding the input. This allows
            for greater flexibility of the ``tile_shape`` but potentially wastes
            more computation (the padded region will be passed into the model
            but will later be discarded from the output tensor).
            If ``True``, incompatible shapes will result in an error.
        verbose: If ``True``, report inference speed.
        report_inp_stats

    Examples:
        >>> model = nn.Sequential(
        ...     nn.Conv2d(5, 32, 3, padding=1), nn.ReLU(),
        ...     nn.Conv2d(32, 2, 1))
        >>> inp = np.random.randn(2, 5, 10, 10)
        >>> predictor = Predictor(model)
        >>> out = predictor.predict(inp)
        >>> assert np.all(np.array(out.shape) == np.array([2, 2, 10, 10]))
    """
    def __init__(
            self,
            model: Union[nn.Module, str, Path],
            state_dict_src: Optional[Union[str, dict]] = None,
            device: Optional[Union[torch.device, str]] = None,
            batch_size: Optional[int] = None,
            tile_shape: Optional[Tuple[int, ...]] = None,
            overlap_shape: Optional[Tuple[int, ...]] = None,
            offset: Optional[Tuple[int, ...]] = None,
            out_shape: Optional[Tuple[int, ...]] = None,
            out_dtype: Optional[torch.dtype] = None,
            float16: bool = False,
            apply_softmax: bool = True,
            transform: Optional[Transform] = None,
            augmentations: Union[int, Optional[Sequence]] = None,
            strict_shapes: bool = False,
            apply_argmax: bool = False,
            argmax_with_threshold: Optional[float] = None,
            verbose: bool = False,
            report_inp_stats: bool = False
    ):
        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
            logger.info(f'Running on device {device}')
        elif isinstance(device, str):
            device = torch.device(device)
        self.device = device
        self.batch_size = batch_size

        self.out_dtype = out_dtype
        self.float16 = float16

        if isinstance(model, Path):
            model = str(model)
        if float16 and not isinstance(model, str) and not next(model.parameters()).dtype == torch.float16:
            # If the model is passed as an object and not already in float16,
            # we need to deepcopy it because model casting to float16 is only
            # supported in-place - thus we would downcast it irreversibly also
            # in the calling scope.
            model = copy.deepcopy(model)
        self.dtype = torch.float16 if float16 else torch.float32
        self.transform = transform
        if isinstance(augmentations, int):
            augmentations = DEFAULT_AUGMENTATIONS_3D[:augmentations]
        self.augmentations = augmentations
        self.strict_shapes = strict_shapes
        self.apply_argmax = apply_argmax
        self.argmax_with_threshold = argmax_with_threshold
        self.verbose = verbose
        self.report_inp_stats = report_inp_stats
        if isinstance(model, str):
            if os.path.isfile(model):
                if model.endswith('.pts'):
                    model = torch.jit.load(model, map_location=device)
                elif model.endswith('.pt'):
                    model = torch.load(model, map_location=device)
                else:
                    raise ValueError(f'{model} has an unkown file extension. Supported are .pt and .pts')
            else:
                raise ValueError(f'Model path {model} not found.')
        self.model = model
        if isinstance(state_dict_src, str):
            state_dict = torch.load(state_dict_src)
            if 'model_state_dict' in state_dict:  # Handle nested dicts
                state_dict = state_dict['model_state_dict']
        elif isinstance(state_dict_src, dict) or state_dict_src is None:
            state_dict = state_dict_src
        else:
            raise ValueError(
                '"state_dict_src" has to be either a path to a .pth file (str),'
                ' a state_dict object (dict) or None.')
        if state_dict is not None:
            set_state_dict(model, state_dict)
        if not apply_softmax and augmentations is not None:
            raise ValueError('When augmentations are enabled, apply_softmax cannot be False.')
        if apply_softmax or augmentations is not None:
            self.model = nn.Sequential(self.model, nn.Softmax(1))
        if float16:
            self.model.half()  # This is destructive. float32 params are lost!
        self.apply_argmax_after_tta = False
        if apply_argmax or argmax_with_threshold is not None:
            self.apply_argmax_after_tta = augmentations is not None
            if not self.apply_argmax_after_tta: # if augmentations are enabled, argmax is applied after augmentations, see _predict
                argmax_layers = [Argmax(dim=1, unsqueeze=True)]
                if argmax_with_threshold:
                    argmax_layers = [nn.Threshold(argmax_with_threshold, 0)] + argmax_layers
                self.model = nn.Sequential(self.model, *argmax_layers)
            if self.out_dtype is None:
                self.out_dtype = torch.uint8
        self._warn_about_shapes = True
        self.model.eval()

        def is_set(array: Tuple[int, ...]):
            return array is not None and np.any(array)

        if is_set(overlap_shape) and is_set(offset):
            raise ValueError(
                f'overlap_shape={overlap_shape} and offet={offset} are both specified, but this is not supported.\n'
                'Either specify overlap_shape (if the spatial shape of inputs and outputs are the same)\n'
                'or offset (if the output is smaller).'
            )

        if not is_set(tile_shape): # no tiling
            assert not (is_set(out_shape) or is_set(overlap_shape) or is_set(offset)), 'If tile_shape is not set, out_shape, overlap_shape and offset should not be set either.'
            self.enable_tiling = False
        else:
            assert is_set(out_shape), 'If tile_shape is set, out_shape is required to be set, too.'
            self.enable_tiling = True
            if offset is None:
                logger.warning(
                    'Predictor: offset=None -> Estimating offset from forward pass. This can fail or silently lead to incorrect results '
                    'and costs an additional forward pass. To avoid this, please set Predictor offset explicitly.'
                )
                offset = utils.calculate_offset(self.model)
            if np.count_nonzero(offset) == 0: # no valid conv → disable offset
                offset = None
            else:
                offset = np.array(offset)
                # Set overlap to offset shape because IMO that's the only reasonable choice.
                overlap_shape = offset
                out_shape = np.array([*out_shape[:-len(offset)], *(out_shape[-len(offset):] - 2 * offset)])
                logger.info(f'Adjusted out_shape: {out_shape}')
        self.offset = offset

        self.overlap_shape = np.array(overlap_shape) if overlap_shape is not None else None
        self.tile_shape = np.array(tile_shape) if tile_shape is not None else None
        self.out_shape = np.array(out_shape) if out_shape is not None else None

    @torch.no_grad()
    def _predict(self, dinp: torch.Tensor, crop_slice=None) -> torch.Tensor:
        dinp = dinp.to(self.device, dtype=self.dtype)
        dout = self.model(dinp)
        if crop_slice is not None:
            dout = dout[crop_slice]

        # Else, apply test-time augmentations and take the mean value.
        # Augmentations are applied directly on the compute device and
        #  intermediate results are stored on-device, so this can increase
        #  GPU memory usage!
        if self.augmentations is not None:
            douts = [dout]
            for aug in self.augmentations:
                dinp_aug = aug.forward(dinp)
                dout_aug = self.model(dinp_aug.to(self.device))
                dout = aug.backward(dout_aug)
                if crop_slice:
                    dout = dout[crop_slice]
                douts.append(dout)
            douts = torch.stack(douts)
            dout = torch.mean(douts, dim=0)

        # if no augmentations, argmax was already applied by the model
        if self.apply_argmax_after_tta:
            if self.argmax_with_threshold:
                dout[dout <= self.argmax_with_threshold] = 0
            dout = dout.argmax(dim=1).to(self.out_dtype)
        dout = dout.to(self.out_dtype)
        return dout

    def _tiled_predict(
            self,
            inp: torch.Tensor,
            out_shape: Optional[Tuple[int]] = None
    ) -> torch.Tensor:
        """Tiled inference with overlapping input tiles.

        Tiling is not used if ``tile_shape`` and ``overlap_shape`` are
        undefined."""
        if self.enable_tiling:
            if self.out_shape is None:
                raise ValueError('If you use tiling, you also need to supply out_shape.')
            out_shape = (inp.shape[0], *out_shape)
            return tiled_apply(
                self._predict,
                inp=inp,
                tile_shape=self.tile_shape,
                overlap_shape=self.overlap_shape,
                offset=self.offset,
                out_shape=out_shape,
                verbose=self.verbose
            )
        # Otherwise: No tiling, apply model to the whole input in one step
        return self._predict(inp)

    def _splitbatch_predict(
            self,
            inp: torch.Tensor,
            num_batches: int,
            out_shape: Optional[Tuple[int]] = None
    ) -> torch.Tensor:
        """Split the input batch into smaller batches of the specified
        ``batch_size`` and perform inference on each of them separately."""
        if self.out_shape is None:
            raise ValueError('If you define a batch_size, you also need to supply out_shape.')
        out = torch.empty((inp.shape[0], *self.out_shape), dtype=self.dtype)
        for k in range(0, num_batches):
            low = self.batch_size * k
            high = self.batch_size * (k + 1)
            out[low:high] = self._tiled_predict(inp[low:high], out_shape=out_shape)
        return out

    def predict(
            self,
            inp: Union[np.ndarray, torch.Tensor],
    ) -> torch.Tensor:
        """ Perform prediction on ``inp`` and return prediction.

        Args:
            inp: Input data, e.g. of shape (N, C, H, W).
                Can be an ``np.ndarray`` or a ``torch.Tensor``.
                Note that ``inp`` is automatically converted to
                the specified ``dtype`` (default: ``torch.float32``) before
                inference.

        Returns:
            Model output
        """
        if self.report_inp_stats:
            from elektronn3.data import utils
            try:
                print('input dist', utils.calculate_means(inp.numpy()), utils.calculate_stds(inp.numpy()))
            except:
                print('input dist', utils.calculate_means(inp), utils.calculate_stds(inp))
        if self.transform is not None:
            if isinstance(inp, torch.Tensor):
                inp = inp.numpy()  # transforms currently only work with numpy ndarrays as in/output
            transformed = np.empty_like(inp)
            for i in range(inp.shape[0]):  # Apply transform for each sample of the batch separately
                transformed[i], _ = self.transform(inp[i], None)  # target=None because we don't have any here
            inp = transformed
        if self.verbose:
            start = time.time()
        # Check/change out_shape for divisibility by tile_shape
        if self.enable_tiling:
            inp, out_shape, relevant_slice = self._ensure_matching_shapes(inp)
        else:
            relevant_slice = None
            out_shape = self.out_shape
        inp = torch.as_tensor(inp, dtype=self.dtype).contiguous()
        inp_batch_size = inp.shape[0]
        spatial_shape = np.array(inp.shape[2:])
        # Lazily figure out these Predictor options based on the input it
        #  receives if they are not already set.
        # Not sure if that's a good idea because these are object-changing
        #  side-effects in the otherwise pure predict() function.
        if self.out_dtype is None:
            self.out_dtype = torch.uint8 if self.argmax_with_threshold is not None else inp.dtype
        if out_shape is not None and out_shape[0] > 255 and self.out_dtype == torch.uint8:
            raise ValueError(f'C = out_shape[0] = {out_shape[0]}, but out_dtype torch.uint8 can only hold values up to 255.')
        if self.tile_shape is None:
            self.tile_shape = spatial_shape
        if self.overlap_shape is None:
            self.overlap_shape = np.zeros_like(spatial_shape)
        if self.batch_size is None:
            self.batch_size = inp_batch_size
        num_batches = int(np.ceil(inp_batch_size / self.batch_size))
        if num_batches == 1:  # Predict everything in one step
            out = self._tiled_predict(inp=inp, out_shape=out_shape)
        else:  # Split input batch into smaller batches and predict separately
            out = self._splitbatch_predict(inp=inp, num_batches=num_batches, out_shape=out_shape)

        # Explicit synchronization so the next GPU operation won't be
        #  mysteriously slow. If we don't synchronize manually here, profilers
        #  will misleadingly report a huge amount of time spent in out.cpu()
        if self.device.type == 'cuda':
            torch.cuda.synchronize()
        out = out.cpu() if relevant_slice is None else out[relevant_slice].cpu()
        if self.verbose:
            dtime = time.time() - start
            amount = out.numel()
            if out_shape is not None and np.array_equal(out_shape[2:], inp.shape[2:]): # calculate the amount of valid data produced
                amount = np.prod([*out.shape[:-3], *(out.shape[-3:] - 2 * self.overlap_shape)])
            speed = amount / dtime / 1e6
            print(f'Inference speed: {speed:.2f} MVox/s, time: {dtime:.2f}.')
        return out

    # TODO: Make this work with input shape != output shape
    def _ensure_matching_shapes(self, inp: np.ndarray) -> Tuple[np.ndarray, Optional[Tuple[int]], Optional[slice]]:
        if self.out_shape is not None and np.any(self.out_shape[1:] % self.tile_shape):
            if self.strict_shapes:
                raise ValueError(
                    'Make sure that out_shape is divisible by tile_shape or '
                    'relax this constraint by setting strict_shapes=False.'
                )
            else:
                padded_out_shape = np.array(self.out_shape)
                padded_out_shape[1:] = np.ceil(self.out_shape[1:] / self.tile_shape) * self.tile_shape
                if self.offset is None:
                    offset = np.zeros(shape=len(padded_out_shape) - 1, dtype=np.int64)
                else:
                    offset = np.array(self.offset)
                padded_inp_shape = (*inp.shape[:2], *padded_out_shape[1:] + 2 * offset)
                padded_inp = np.zeros(padded_inp_shape)
                # Define the relevant region (that is: without the padding that was just added)
                relevant_slice_inp = _extend_nc([slice(0, d) for d in inp.shape[2:]])
                relevant_slice_out = _extend_nc([slice(0, d) for d in self.out_shape[1:]])
                padded_inp[relevant_slice_inp] = inp
                if self._warn_about_shapes and np.any(padded_out_shape != self.out_shape):
                    sh_diff = np.subtract(padded_out_shape, self.out_shape)
                    # Only nonzero elements are multiplied, otherwise it will be 0.
                    wasted_pix = np.prod(sh_diff[sh_diff != 0])
                    total_pix = np.prod(padded_out_shape)
                    wasted_percentage = 100 * wasted_pix / total_pix
                    logger.info(
                        f'Adapting out_shape {tuple(self.out_shape[1:])} to '
                        f'tile_shape {tuple(self.tile_shape)} '
                        f'by padding out_shape to {tuple(padded_out_shape[1:])}.\n'
                        f'Suboptimal shapes will reduce execution speed.'
                        # f'At least {wasted_percentage:.2f}% of total compute will be '
                        # f'wasted by this padding.'
                    )
                    self._warn_about_shapes = False
                    # TODO: Calculate exact compute waste by looking at increased tile overlaps
                    #  (the current estimation omits the (potentially high-impact) added per-tile
                    #  padding/overlaps via overlap_shape.
        else:
            padded_inp = inp
            padded_out_shape = self.out_shape
            relevant_slice_out = None
        return padded_inp, padded_out_shape, relevant_slice_out

    def predict_proba(self, inp):
        logger.warning('Predictor.predict_proba(inp) is deprecated. Please use Predictor.predict(inp) instead.')
        return self.predict(inp)


# TODO: This can be replaced with a single model.load_state_dict(state_dict) call
#       after a while, because Trainer._save_model() now always saves unwrapped
#       modules if a parallel wrapper is detected. Or should we still keep this
#       for better support of models accidentally saved in wrapped state?
def set_state_dict(model: torch.nn.Module, state_dict: dict):
    """Set state dict of a model.

    Also works with ``torch.nn.DataParallel`` models."""
    try:
        model.load_state_dict(state_dict)
    # If self.model was saved as nn.DataParallel then remove 'module.' prefix
    # in every key
    except RuntimeError:  # TODO: Is it safe to catch all runtime errors here?
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            new_state_dict[k.replace('module.', '')] = v
        model.load_state_dict(new_state_dict)



# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Maxim Berman, Martin Drawitsch

# This file is mostly copied from https://github.com/bermanmaxim/LovaszSoftmax.
# Modifications for elektronn3:
# - Support 5D images
# - Avoid zero division in low-precision training by adding a small constant eps in divisions

"""
Lovasz-Softmax and Jaccard hinge loss in PyTorch
Maxim Berman 2018 ESAT-PSI KU Leuven (MIT License)

"""

from __future__ import print_function, division

import torch
from torch.autograd import Variable
import torch.nn.functional as F
import numpy as np
try:
    from itertools import  ifilterfalse
except ImportError: # py3k
    from itertools import  filterfalse


eps = 0.0001  # To avoid divisions by zero, esp. in low-precision training


def lovasz_grad(gt_sorted):
    """
    Computes gradient of the Lovasz extension w.r.t sorted errors
    See Alg. 1 in paper
    """
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1. - intersection / (union + eps)
    if p > 1: # cover 1-pixel case
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard


def iou_binary(preds, labels, EMPTY=1., ignore=None, per_image=True):
    """
    IoU for foreground class
    binary: 1 foreground, 0 background
    """
    if not per_image:
        preds, labels = (preds,), (labels,)
    ious = []
    for pred, label in zip(preds, labels):
        intersection = ((label == 1) & (pred == 1)).sum()
        union = ((label == 1) | ((pred == 1) & (label != ignore))).sum()
        if not union:
            iou = EMPTY
        else:
            iou = float(intersection) / (union + eps)
        ious.append(iou)
    iou = mean(ious)    # mean accross images if per_image
    return 100 * iou


def iou(preds, labels, C, EMPTY=1., ignore=None, per_image=False):
    """
    Array of IoU for each (non ignored) class
    """
    if not per_image:
        preds, labels = (preds,), (labels,)
    ious = []
    for pred, label in zip(preds, labels):
        iou = []    
        for i in range(C):
            if i != ignore: # The ignored label is sometimes among predicted classes (ENet - CityScapes)
                intersection = ((label == i) & (pred == i)).sum()
                union = ((label == i) | ((pred == i) & (label != ignore))).sum()
                if not union:
                    iou.append(EMPTY)
                else:
                    iou.append(float(intersection) / (union + eps))
        ious.append(iou)
    ious = map(mean, zip(*ious)) # mean accross images if per_image
    return 100 * np.array(ious)


# --------------------------- BINARY LOSSES ---------------------------


def lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
      logits: [B, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      per_image: compute the loss per image instead of per batch
      ignore: void class id
    """
    if per_image:
        loss = mean(lovasz_hinge_flat(*flatten_binary_scores(log.unsqueeze(0), lab.unsqueeze(0), ignore))
                          for log, lab in zip(logits, labels))
    else:
        loss = lovasz_hinge_flat(*flatten_binary_scores(logits, labels, ignore))
    return loss


def lovasz_hinge_flat(logits, labels):
    """
    Binary Lovasz hinge loss
      logits: [P] Variable, logits at each prediction (between -\infty and +\infty)
      labels: [P] Tensor, binary ground truth labels (0 or 1)
      ignore: label to ignore
    """
    if len(labels) == 0:
        # only void pixels, the gradients should be 0
        return logits.sum() * 0.
    signs = 2. * labels.float() - 1.
    errors = (1. - logits * Variable(signs))
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def flatten_binary_scores(scores, labels, ignore=None):
    """
    Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = (labels != ignore)
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


class StableBCELoss(torch.nn.modules.Module):
    def __init__(self):
         super(StableBCELoss, self).__init__()
    def forward(self, input, target):
         neg_abs = - input.abs()
         loss = input.clamp(min=0) - input * target + (1 + neg_abs.exp()).log()
         return loss.mean()


def binary_xloss(logits, labels, ignore=None):
    """
    Binary Cross entropy loss
      logits: [B, H, W] Variable, logits at each pixel (between -\infty and +\infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      ignore: void class id
    """
    logits, labels = flatten_binary_scores(logits, labels, ignore)
    loss = StableBCELoss()(logits, Variable(labels.float()))
    return loss


# --------------------------- MULTICLASS LOSSES ---------------------------


def lovasz_softmax(probas, labels, only_present=False, per_image=False, ignore=None):
    """
    Multi-class Lovasz-Softmax loss
      probas: [B, C, H, W] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [B, H, W] Tensor, ground truth labels (between 0 and C - 1)
      only_present: average only on num_classes present in ground truth
      per_image: compute the loss per image instead of per batch
      ignore: void class labels
    """
    if per_image:
        loss = mean(lovasz_softmax_flat(*flatten_probas(prob.unsqueeze(0), lab.unsqueeze(0), ignore), only_present=only_present)
                          for prob, lab in zip(probas, labels))
    else:
        loss = lovasz_softmax_flat(*flatten_probas(probas, labels, ignore), only_present=only_present)
    return loss


def lovasz_softmax_flat(probas, labels, only_present=False):
    """
    Multi-class Lovasz-Softmax loss
      probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [P] Tensor, ground truth labels (between 0 and C - 1)
      only_present: average only on num_classes present in ground truth
    """
    C = probas.size(1)
    losses = []
    for c in range(C):
        fg = (labels == c).float() # foreground for class c
        if only_present and fg.sum() == 0:
            continue
        errors = (Variable(fg) - probas[:, c]).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(lovasz_grad(fg_sorted))))
    return mean(losses)


def flatten_probas(probas, labels, ignore=None):
    """
    Flattens predictions in the batch
    """
    C = probas.shape[1]
    if probas.dim() == 4:
        probas = probas.permute(0, 2, 3, 1).contiguous().view(-1, C)  # B * H * W, C = P, C
    elif probas.dim() == 5:  # 3D images
        probas = probas.permute(0, 2, 3, 4, 1).contiguous().view(-1, C)  # B * D * H * W, C = P, C
    labels = labels.view(-1)
    if ignore is None:
        return probas, labels
    valid = (labels != ignore)
    vprobas = probas[valid.nonzero().squeeze()]
    vlabels = labels[valid]
    return vprobas, vlabels

def xloss(logits, labels, ignore=None):
    """
    Cross entropy loss
    """
    return F.cross_entropy(logits, Variable(labels), ignore_index=255)


# --------------------------- HELPER FUNCTIONS ---------------------------

def mean(l, ignore_nan=False, empty=0):
    """
    nanmean compatible with generators.
    """
    l = iter(l)
    if ignore_nan:
        l = ifilterfalse(np.isnan, l)
    try:
        n = 1
        acc = next(l)
    except StopIteration:
        if empty == 'raise':
            raise ValueError('Empty mean')
        return empty
    for n, v in enumerate(l, 2):
        acc += v
    if n == 1:
        return acc
    return acc / n



"""Code related to data sources (HDF5 etc.)"""

# ELEKTRONN3 - Neural Network Toolkit
#
# Copyright (c) 2017 - now
# Max Planck Institute of Neurobiology, Munich, Germany
# Authors: Martin Drawitsch

import os
from typing import Union, Any, Sequence

import h5py
import numpy as np


class DataSource:  #(Protocol):  # Protocol requires Python 3.8 or typing_extensions...
    def __getitem__(self, idx: Union[int, slice]) -> np.ndarray: ...

    # Expected properties: size, shape, dtype, fname, in_memory, ndim


class HDF5DataSource(DataSource):
    """An h5py.Dataset wrapper for safe multiprocessing. Opens the file and
    the dataset on each read/property access and then immediately closes it.

    This is a workaround for this issue and related data corruptions:
    https://github.com/pytorch/pytorch/issues/11929.

    By avoiding open file handles before worker processes are forked,
    concurrency issues with HDF5's global state do not apply."""

    def __init__(self, fname: str, key: str, in_memory: bool = False):
        self.fname = os.path.expanduser(fname)
        self.key = key
        self.in_memory = in_memory

        if self.in_memory:
            self._data: np.ndarray
            self._initialize_memory()

    def _initialize_memory(self) -> None:
        with h5py.File(self.fname, 'r') as f:
            h5data = f[self.key]
            self._data = h5data[()]

    # Wraps direct attribute, property and method access
    def __getattr__(self, attr: str) -> Any:
        if self.in_memory:
            h5data = self._data
            return getattr(h5data, attr)
        with h5py.File(self.fname, 'r') as f:
            h5data = f[self.key]
            return getattr(h5data, attr)

    # But dunder methods have to be wrapped manually: https://stackoverflow.com/a/3700899
    def __getitem__(self, idx: Union[int, slice]) -> np.ndarray:
        if self.in_memory:
            h5data = self._data
            return h5data[idx]
        with h5py.File(self.fname, 'r') as f:
            h5data = f[self.key]
            return h5data[idx]


def slice_3d(
        src: DataSource,
        coords_lo: Sequence[int],
        coords_hi: Sequence[int],
        dtype: type = np.float32,
        prepend_empty_axis: bool = False,
        check_bounds=True,
) -> np.ndarray:
    """ Slice a patch of 3D image data out of a data source.

    Args:
        src: Source data set from which to read data.
            The expected data shapes are (C, D, H, W) or (D, H, W).
        coords_lo: Lower bound of the coordinates where data should be read
            from in ``src``.
        coords_hi: Upper bound of the coordinates where data should be read
            from in ``src``.
        dtype: NumPy ``dtype`` that the sliced array will be cast to if it
            doesn't already have this dtype.
        prepend_empty_axis: Prepends a new empty (1-sized) axis to the sliced
            array before returning it.
        check_bounds: If ``True`` (default), only indices that are within the
            bounds of ``src`` will be allowed (no negative indices or slices
            to indices that exceed the shape of ``src``, which would normally
            just be ignored).

    Returns:
        Sliced image array.
    """
    if check_bounds:
        if np.any(np.array(coords_lo) < 0):
            raise RuntimeError(f'coords_lo={coords_lo} exceeds src shape {src.shape[-3:]}')
        if np.any(np.array(coords_hi) > np.array(src.shape[-3:])):
            raise RuntimeError(f'coords_hi={coords_hi} exceeds src shape {src.shape[-3:]}')

    # Generalized n-d slicing code (temporarily disabled because of the
    #  performance issue described in the comment below):
    ## full_slice = calculate_nd_slice(src, coords_lo, coords_hi)
    ## # # TODO: Use a better workaround or fix this in h5py:
    ## srcv = src.value  # Workaround for hp5y indexing limitation. The `.value` call is very unfortunate! It loads the entire cube to RAM.
    ## cut = srcv[full_slice]

    if src.ndim == 4:
        cut = src[
            :,
            coords_lo[0]:coords_hi[0],
            coords_lo[1]:coords_hi[1],
            coords_lo[2]:coords_hi[2]
        ]
    elif src.ndim == 3:
        cut = src[
            coords_lo[0]:coords_hi[0],
            coords_lo[1]:coords_hi[1],
            coords_lo[2]:coords_hi[2]
        ]
    else:
        raise ValueError(f'Expected src.ndim to be 3 or 4, but got {src.ndim} instead.')
    if prepend_empty_axis:
        cut = cut[None]
    cut = cut.astype(dtype, copy=False)
    return cut
